##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing TINKER, implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
"""
import glob
import os
import tempfile

from distutils.version import LooseVersion

import easybuild.tools.toolchain as toolchain
from easybuild.framework.easyblock import EasyBlock
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import apply_regex_substitutions, change_dir, copy_dir, copy_file, mkdir
from easybuild.tools.modules import get_software_root
from easybuild.tools.run import run_cmd
from easybuild.tools.systemtools import DARWIN, LINUX, get_os_type


class EB_TINKER(EasyBlock):
    """Support for building/installing TINKER."""

    def __init__(self, *args, **kwargs):
        """Custom easyblock constructor for TINKER: initialise class variables."""
        super(EB_TINKER, self).__init__(*args, **kwargs)

        self.build_subdir = None
        self.build_in_installdir = True

    def configure_step(self):
        """Custom configuration procedure for TINKER."""
        # make sure FFTW is available
        if get_software_root('FFTW') is None:
            raise EasyBuildError("FFTW dependency is not available.")

        os_dirs = {
            LINUX: 'linux',
            DARWIN: 'macosx',
        }
        os_type = get_os_type()
        os_dir = os_dirs.get(os_type)
        if os_dir is None:
            raise EasyBuildError("Failed to determine OS directory for %s (known: %s)", os_type, os_dirs)

        comp_dirs = {
            toolchain.INTELCOMP: 'intel',
            toolchain.GCC: 'gfortran',
        }
        comp_fam = self.toolchain.comp_family()
        comp_dir = comp_dirs.get(comp_fam)
        if comp_dir is None:
            raise EasyBuildError("Failed to determine compiler directory for %s (known: %s)", comp_fam, comp_dirs)

        self.build_subdir = os.path.join(os_dir, comp_dir)
        self.log.info("Using build scripts from %s subdirectory" % self.build_subdir)

        # patch 'link.make' script to use FFTW provided via EasyBuild
        link_make_fp = os.path.join(self.cfg['start_dir'], self.build_subdir, 'link.make')
        regex_subs = [(r"libfftw3_threads.a libfftw3.a", r"-L$EBROOTFFTW/lib -lfftw3_omp -lfftw3")]
        apply_regex_substitutions(link_make_fp, regex_subs)

        # patch *.make files to get rid of hardcoded -openmp flag,
        # which doesn't work anymore with recent Intel compilers
        if comp_fam == toolchain.INTELCOMP:
            make_fps = glob.glob(os.path.join(self.cfg['start_dir'], self.build_subdir, '*.make'))
            regex_subs = [(r'-openmp', r'-fopenmp')]
            for make_fp in make_fps:
                apply_regex_substitutions(make_fps, regex_subs)

    def build_step(self):
        """Custom build procedure for TINKER."""

        change_dir(os.path.join(self.cfg['start_dir'], 'source'))

        run_cmd(os.path.join(self.cfg['start_dir'], self.build_subdir, 'compile.make'))
        run_cmd(os.path.join(self.cfg['start_dir'], self.build_subdir, 'library.make'))
        run_cmd(os.path.join(self.cfg['start_dir'], self.build_subdir, 'link.make'))

    def test_step(self):
        """Custom built-in test procedure for TINKER."""
        if self.cfg['runtest']:
            # copy tests, params and built binaries to temporary directory for testing
            tmpdir = tempfile.mkdtemp()
            testdir = os.path.join(tmpdir, 'test')

            mkdir(os.path.join(tmpdir, 'bin'))
            binaries = glob.glob(os.path.join(self.cfg['start_dir'], 'source', '*.x'))
            for binary in binaries:
                copy_file(binary, os.path.join(tmpdir, 'bin', os.path.basename(binary)[:-2]))
            copy_dir(os.path.join(self.cfg['start_dir'], 'test'), testdir)
            copy_dir(os.path.join(self.cfg['start_dir'], 'params'), os.path.join(tmpdir, 'params'))

            change_dir(testdir)

            # run all tests via the provided 'run' scripts
            tests = glob.glob(os.path.join(testdir, '*.run'))

            # gpcr takes too long (~1h)
            skip_tests = ['gpcr']
            if (LooseVersion(self.version) < LooseVersion('8.7.2')):
                # ifabp fails due to input issues (?)
                skip_tests.append('ifabp')
            if (LooseVersion(self.version) >= LooseVersion('8.7.2')):
                # salt and dialinine takes too long
                skip_tests.extend(['salt', 'dialanine'])

            tests = [t for t in tests if not any([t.endswith('%s.run' % x) for x in skip_tests])]

            for test in tests:
                run_cmd(test)

    def install_step(self):
        """Custom install procedure for TINKER."""

        change_dir(os.path.join(self.cfg['start_dir'], 'source'))

        mkdir(os.path.join(self.cfg['start_dir'], 'bin'))
        run_cmd(os.path.join(self.cfg['start_dir'], self.build_subdir, 'rename.make'))

    def sanity_check_step(self):
        """Custom sanity check for TINKER."""
        custom_paths = {
            'files': ['tinker/source/libtinker.a'],
            'dirs': ['tinker/bin'],
        }
        super(EB_TINKER, self).sanity_check_step(custom_paths=custom_paths)

    def make_module_req_guess(self):
        """Custom guesses for module file prepend-path statements."""
        guesses = super(EB_TINKER, self).make_module_req_guess()
        guesses['PATH'].append(os.path.join('tinker', 'bin'))
        guesses['LIBRARY_PATH'].append(os.path.join('tinker', 'source'))
        return guesses



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing scipy, implemented as an easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
@author: Jasper Grimm (University of York)
"""
import os
import tempfile
from distutils.version import LooseVersion

import easybuild.tools.environment as env
import easybuild.tools.toolchain as toolchain
from easybuild.easyblocks.generic.fortranpythonpackage import FortranPythonPackage
from easybuild.easyblocks.generic.mesonninja import MesonNinja
from easybuild.easyblocks.generic.pythonpackage import PythonPackage, det_pylibdir
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError, print_warning
from easybuild.tools.filetools import change_dir, copy_dir, copy_file


class EB_scipy(FortranPythonPackage, PythonPackage, MesonNinja):
    """Support for installing the scipy Python package as part of a Python installation."""

    @staticmethod
    def extra_options():
        """Easyconfig parameters specific to scipy."""
        extra_vars = ({
            'enable_slow_tests': [False, "Run scipy test suite, including tests marked as slow", CUSTOM],
            'ignore_test_result': [None, "Run scipy test suite, but ignore test failures (True/False/None). Default "
                                         "(None) implies True for scipy < 1.9, and False for scipy >= 1.9", CUSTOM],
        })

        return PythonPackage.extra_options(extra_vars=extra_vars)

    def __init__(self, *args, **kwargs):
        """Set scipy-specific test command."""
        # calling PythonPackage __init__ also lets MesonNinja work in an extension
        PythonPackage.__init__(self, *args, **kwargs)
        self.testinstall = True

        if LooseVersion(self.version) >= LooseVersion('1.9'):
            self.use_meson = True

            # enforce scipy test suite results if not explicitly disabled for scipy >= 1.9
            # strip inherited PythonPackage installopts
            installopts = self.cfg['installopts']
            pythonpackage_installopts = ['--no-deps', '--ignore-installed', '--no-index', '--egg',
                                         '--zip-ok', '--no-index']
            self.log.info("Stripping inherited PythonPackage installopts %s from installopts %s",
                          pythonpackage_installopts, installopts)
            for i in pythonpackage_installopts:
                installopts = installopts.replace(i, '')
            self.cfg['installopts'] = installopts

        else:
            self.use_meson = False

        if self.cfg['ignore_test_result'] is None:
            # automatically ignore scipy test suite results for scipy < 1.9, as we did in older easyconfigs
            self.cfg['ignore_test_result'] = LooseVersion(self.version) < '1.9'
            self.log.info("ignore_test_result not specified, so automatically set to %s for scipy %s",
                          self.cfg['ignore_test_result'], self.version)

        if self.cfg['ignore_test_result']:
            # used to maintain compatibility with easyconfigs predating scipy 1.9;
            # runs tests (serially) in a way that exits with code 0 regardless of test results,
            # see https://github.com/easybuilders/easybuild-easyblocks/issues/2237
            self.testcmd = "cd .. && %(python)s -c 'import numpy; import scipy; scipy.test(verbose=2)'"
        else:
            self.testcmd = " && ".join([
                "cd ..",
                "touch %(srcdir)s/.coveragerc",
                "%(python)s %(srcdir)s/runtests.py -v --no-build --parallel %(parallel)s",
            ])
            if self.cfg['enable_slow_tests']:
                self.testcmd += " -m full "

    def configure_step(self):
        """Custom configure step for scipy: set extra installation options when needed."""

        # scipy >= 1.9.0 uses Meson/Ninja
        if self.use_meson:
            # configure BLAS/LAPACK library to use with Meson for scipy >= 1.9.0
            lapack_lib = self.toolchain.lapack_family()
            if lapack_lib == toolchain.FLEXIBLAS:
                blas_lapack = 'flexiblas'
            elif lapack_lib == toolchain.INTELMKL:
                blas_lapack = 'mkl'
            elif lapack_lib == toolchain.OPENBLAS:
                blas_lapack = 'openblas'
            else:
                raise EasyBuildError("Unknown BLAS/LAPACK library used: %s", lapack_lib)

            for opt in ('blas', 'lapack'):
                self.cfg.update('configopts', '-D%(opt)s=%(blas_lapack)s' % {'opt': opt, 'blas_lapack': blas_lapack})

            # need to have already installed extensions in PATH, PYTHONPATH for configure/build/install steps
            pythonpath = os.getenv('PYTHONPATH')
            pylibdir = det_pylibdir()
            env.setvar('PYTHONPATH', os.pathsep.join([os.path.join(self.installdir, pylibdir), pythonpath]))

            path = os.getenv('PATH')
            env.setvar('PATH', os.pathsep.join([os.path.join(self.installdir, 'bin'), path]))

            MesonNinja.configure_step(self)

        else:
            # scipy < 1.9.0 uses install procedure using setup.py
            FortranPythonPackage.configure_step(self)

        if LooseVersion(self.version) >= LooseVersion('0.13'):
            # in recent scipy versions, additional compilation is done in the install step,
            # which requires unsetting $LDFLAGS
            if self.toolchain.comp_family() in [toolchain.GCC, toolchain.CLANGGCC]:  # @UndefinedVariable
                self.cfg.update('preinstallopts', "unset LDFLAGS && ")

    def build_step(self):
        """Custom build step for scipy: use ninja for scipy >= 1.9.0"""
        if self.use_meson:
            MesonNinja.build_step(self)
        else:
            FortranPythonPackage.build_step(self)

    def test_step(self):
        """Run available scipy unit tests. Adapted from numpy easyblock"""

        if self.use_meson:
            # temporarily install scipy so we can run the test suite
            tmpdir = tempfile.mkdtemp()
            cwd = os.getcwd()

            tmp_builddir = os.path.join(tmpdir, 'build')
            tmp_installdir = os.path.join(tmpdir, 'install')

            # create a copy of the builddir
            copy_dir(cwd, tmp_builddir)
            change_dir(tmp_builddir)

            # reconfigure (to update prefix), and install to tmpdir
            MesonNinja.configure_step(self, cmd_prefix=tmp_installdir)
            MesonNinja.install_step(self)

            tmp_pylibdir = [os.path.join(tmp_installdir, det_pylibdir())]
            self.prepare_python()

            self.cfg['pretestopts'] = " && ".join([
                # LDFLAGS should not be set when testing numpy/scipy, because it overwrites whatever numpy/scipy sets
                # see http://projects.scipy.org/numpy/ticket/182
                "unset LDFLAGS",
                "export PYTHONPATH=%s:$PYTHONPATH" % tmp_pylibdir,
                "",
            ])
            self.cfg['runtest'] = self.testcmd % {
                'python': self.python_cmd,
                'srcdir': self.cfg['start_dir'],
                'parallel': self.cfg['parallel'],
            }

            MesonNinja.test_step(self)

        else:
            self.testcmd = self.testcmd % {
                'python': '%(python)s',
                'srcdir': self.cfg['start_dir'],
                'parallel': self.cfg['parallel'],
            }
            FortranPythonPackage.test_step(self)

    def install_step(self):
        """Custom install step for scipy: use ninja for scipy >= 1.9.0"""
        if self.use_meson:
            MesonNinja.install_step(self)

            # copy PKG-INFO file included in scipy source tarball to scipy-<version>.egg-info in installation,
            # so pip is aware of the scipy installation (required for 'pip list', 'pip check', etc.);
            # see also https://github.com/easybuilders/easybuild-easyblocks/issues/2901
            pkg_info = os.path.join(self.cfg['start_dir'], 'PKG-INFO')
            target_egg_info = os.path.join(self.installdir, self.pylibdir, 'scipy-%s.egg-info' % self.version)
            if os.path.isfile(pkg_info):
                copy_file(pkg_info, target_egg_info)
            else:
                cwd = os.getcwd()
                print_warning("%s not found in %s, so can't use it to create %s!", pkg_info, cwd, target_egg_info,
                              log=self.log)
        else:
            FortranPythonPackage.install_step(self)

    def sanity_check_step(self, *args, **kwargs):
        """Custom sanity check for scipy."""

        # can't use self.pylibdir here, need to determine path on the fly using currently active 'python' command;
        # this is important for numpy installations for multiple Python version (via multi_deps)
        custom_paths = {
            'files': [],
            'dirs': [det_pylibdir()],
        }

        # make sure that scipy is included in output of 'pip list',
        # so that 'pip check' passes if scipy is a required dependency for another Python package;
        # use case-insensitive match, since name is sometimes reported as 'SciPy'
        custom_commands = [r"pip list | grep -iE '^scipy\s+%s\s*$'" % self.version.replace('.', r'\.')]

        return PythonPackage.sanity_check_step(self, custom_paths=custom_paths, custom_commands=custom_commands)



##
# Copyright 2021-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
Support for building and installing AOMP - AMD OpenMP compiler, implemented as
an EasyBlock

@author: Jorgen Nordmoen (University Center for Information Technology - UiO)
"""
import os

from easybuild.easyblocks.generic.binary import Binary
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError, print_warning
from easybuild.tools.config import build_option
from easybuild.tools.filetools import move_file, remove_file
from easybuild.tools.modules import get_software_root
from easybuild.tools.systemtools import AARCH64, POWER, X86_64
from easybuild.tools.systemtools import get_cpu_architecture, get_shared_lib_ext

AOMP_ALL_COMPONENTS = ['roct', 'rocr', 'project', 'libdevice', 'openmp',
                       'extras', 'pgmath', 'flang', 'flang_runtime', 'comgr',
                       'rocminfo', 'vdi', 'hipvdi', 'ocl', 'rocdbgapi',
                       'rocgdb', 'roctracer', 'rocprofiler']
AOMP_DEFAULT_COMPONENTS = ['roct', 'rocr', 'project', 'libdevice', 'openmp',
                           'extras', 'pgmath', 'flang', 'flang_runtime',
                           'comgr', 'rocminfo']
AOMP_X86_COMPONENTS = ['vdi', 'hipvdi', 'ocl']
AOMP_DBG_COMPONENTS = ['rocdbgapi', 'rocgdb']
AOMP_PROF_COMPONENTS = ['roctracer', 'rocprofiler']


class EB_AOMP(Binary):
    """Support for installing AOMP"""

    @staticmethod
    def extra_options():
        extra_vars = Binary.extra_options()
        extra_vars.update({
            'components': [None, "AOMP components to build. Possible components: " +
                           ', '.join(AOMP_ALL_COMPONENTS), CUSTOM],
        })
        return extra_vars

    def __init__(self, *args, **kwargs):
        """Initialize custom class variables for Clang."""
        super(EB_AOMP, self).__init__(*args, **kwargs)
        self.cfg['extract_sources'] = True
        self.cfg['dontcreateinstalldir'] = True

    def configure_step(self):
        """Configure AOMP build and let 'Binary' install"""
        # Setup install command
        self.cfg['install_cmd'] = './aomp/bin/build_aomp.sh'
        # Setup 'preinstallopts'
        version_major = self.version.split('.')[0]
        install_options = [
            'AOMP={!s}'.format(self.installdir),
            'AOMP_REPOS="{!s}/aomp{!s}"'.format(self.builddir, version_major),
            'AOMP_CMAKE={!s}/bin/cmake'.format(get_software_root('CMake')),
            'AOMP_CHECK_GIT_BRANCH=0',
            'AOMP_APPLY_ROCM_PATCHES=0',
            'AOMP_STANDALONE_BUILD=1',
        ]
        if self.cfg['parallel']:
            install_options.append(
                'NUM_THREADS={!s}'.format(self.cfg['parallel']))
        else:
            install_options.append('NUM_THREADS=1')
        # Check if CUDA is loaded and alternatively build CUDA backend
        if get_software_root('CUDA') or get_software_root('CUDAcore'):
            cuda_root = get_software_root('CUDA') or get_software_root('CUDAcore')
            install_options.append('AOMP_BUILD_CUDA=1')
            install_options.append('CUDA="{!s}"'.format(cuda_root))
            # list of CUDA compute capabilities to use can be specifed in two ways (where (2) overrules (1)):
            # (1) in the easyconfig file, via the custom cuda_compute_capabilities;
            # (2) in the EasyBuild configuration, via --cuda-compute-capabilities configuration option;
            ec_cuda_cc = self.cfg['cuda_compute_capabilities']
            cfg_cuda_cc = build_option('cuda_compute_capabilities')
            cuda_cc = cfg_cuda_cc or ec_cuda_cc or []
            if cfg_cuda_cc and ec_cuda_cc:
                warning_msg = "cuda_compute_capabilities specified in easyconfig (%s) are overruled by " % ec_cuda_cc
                warning_msg += "--cuda-compute-capabilities configuration option (%s)" % cfg_cuda_cc
                print_warning(warning_msg)
            if not cuda_cc:
                raise EasyBuildError("CUDA module was loaded, "
                                     "indicating a build with CUDA, "
                                     "but no CUDA compute capability "
                                     "was specified!")
            # Convert '7.0' to '70' format
            cuda_cc = [cc.replace('.', '') for cc in cuda_cc]
            cuda_str = ",".join(cuda_cc)
            install_options.append('NVPTXGPUS="{!s}"'.format(cuda_str))
        else:
            # Explicitly disable CUDA
            install_options.append('AOMP_BUILD_CUDA=0')
        # Combine install instructions above into 'preinstallopts'
        self.cfg['preinstallopts'] = ' '.join(install_options)
        # Setup components for install
        components = self.cfg.get('components', None)
        # If no components were defined we use the default
        if not components:
            components = AOMP_DEFAULT_COMPONENTS
            # NOTE: The following has not been tested properly and is therefore
            # removed
            #
            # Add X86 components if correct architecture
            # if get_cpu_architecture() == X86_64:
            #     components.extend(AOMP_X86_COMPONENTS)
        # Only build selected components
        self.cfg['installopts'] = 'select ' + ' '.join(components)

    def post_install_step(self):
        super(EB_AOMP, self).post_install_step()
        # The install script will create a symbolic link as the install
        # directory, this creates problems for EB as it won't remove the
        # symlink. To remedy this we remove the link here and rename the actual
        # install directory created by the AOMP install script
        if os.path.islink(self.installdir):
            remove_file(self.installdir)
        else:
            err_str = "Expected '{!s}' to be a symbolic link" \
                      " that needed to be removed, but it wasn't!"
            raise EasyBuildError(err_str.format(self.installdir))
        # Move the actual directory containing the install
        install_name = '{!s}_{!s}'.format(os.path.basename(self.installdir),
                                          self.version)
        actual_install = os.path.join(os.path.dirname(self.installdir),
                                      install_name)
        if os.path.exists(actual_install) and os.path.isdir(actual_install):
            move_file(actual_install, self.installdir)
        else:
            err_str = "Tried to move '{!s}' to '{!s}', " \
                      " but it either doesn't exist" \
                      " or isn't a directory!"
            raise EasyBuildError(err_str.format(actual_install,
                                                self.installdir))

    def sanity_check_step(self):
        """Custom sanity check for AOMP"""
        shlib_ext = get_shared_lib_ext()
        arch = get_cpu_architecture()
        # Check architecture explicitly since Clang uses potentially
        # different names
        arch_map = {
            X86_64: 'x86_64',
            POWER: 'ppc64',
            AARCH64: 'aarch64',
        }

        if arch in arch_map:
            arch = arch_map[arch]
        else:
            print_warning("Unknown CPU architecture (%s) for OpenMP offloading!" % arch)
        custom_paths = {
            'files': [
                "amdgcn/bitcode/hip.bc", "amdgcn/bitcode/opencl.bc", "bin/aompcc",
                "bin/aompversion", "bin/clang", "bin/flang", "bin/ld.lld", "bin/llvm-config",
                "bin/mygpu", "bin/opt", "bin/rocminfo", "include/amd_comgr.h",
                "include/hsa/amd_hsa_common.h", "include/hsa/hsa.h", "include/omp.h",
                "include/omp_lib.h", "lib/libclang.%s" % shlib_ext, "lib/libflang.%s" % shlib_ext,
                "lib/libomp.%s" % shlib_ext, "lib/libomptarget.rtl.amdgpu.%s" % shlib_ext,
                "lib/libomptarget.rtl.%s.%s" % (arch, shlib_ext), "lib/libomptarget.%s" % shlib_ext,
            ],
            'dirs': ["amdgcn", "include/clang", "include/hsa", "include/llvm"],
        }
        # If we are building with CUDA support we need to check if it was built properly
        if get_software_root('CUDA') or get_software_root('CUDAcore'):
            custom_paths['files'].append("lib/libomptarget.rtl.cuda.%s" % shlib_ext)
        custom_commands = [
            'aompcc --help', 'clang --help', 'clang++ --help', 'flang --help',
            'llvm-config --cxxflags',
        ]
        super(EB_AOMP, self).sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for ABAQUS, implemented as an easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
@author: Simon Branford (University of Birmingham)
"""
from distutils.version import LooseVersion
import glob
import os

from easybuild.easyblocks.generic.binary import Binary
from easybuild.framework.easyblock import EasyBlock
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.environment import setvar
from easybuild.tools.filetools import change_dir, symlink, write_file
from easybuild.tools.modules import get_software_root
from easybuild.tools.run import run_cmd_qa
from easybuild.tools.systemtools import get_os_name
from easybuild.tools.py2vs3 import OrderedDict


class EB_ABAQUS(Binary):
    """Support for installing ABAQUS."""

    @staticmethod
    def extra_options():
        extra_vars = {
            'with_fe_safe': [None, "Enable installation of fe-safe (if None: auto-enable for ABAQUS >= 2020)", CUSTOM],
            'with_tosca': [None, "Enable installation of Tosca (if None: auto-enable for ABAQUS >= 2020)", CUSTOM],
        }
        return Binary.extra_options(extra_vars)

    def __init__(self, *args, **kwargs):
        """Initialisation of custom class variables for ABAQUS."""
        super(EB_ABAQUS, self).__init__(*args, **kwargs)
        self.replayfile = None

        if self.cfg['with_tosca'] is None and LooseVersion(self.version) >= LooseVersion('2020'):
            self.log.info("Auto-enabling installation of Tosca components for ABAQUS versions >= 2020")
            self.cfg['with_tosca'] = True

        if self.cfg['with_fe_safe'] is None and LooseVersion(self.version) >= LooseVersion('2020'):
            self.log.info("Auto-enabling installation of fe-safe components for ABAQUS versions >= 2020")
            self.cfg['with_fe_safe'] = True

    def extract_step(self):
        """Use default extraction procedure instead of the one for the Binary easyblock."""
        EasyBlock.extract_step(self)

    def configure_step(self):
        """Configure ABAQUS installation."""
        if LooseVersion(self.version) >= LooseVersion('2016'):
            # Rocky Linux isn't recognized; faking it as RHEL
            if get_os_name() in ['Rocky Linux', 'AlmaLinux']:
                setvar('DISTRIB_ID', 'RedHatEnterpriseServer')
            # skip checking of Linux version
            setvar('DSY_Force_OS', 'linux_a64')
            # skip checking of license server
            setvar('NOLICENSECHECK', 'true')
        else:
            self.replayfile = os.path.join(self.builddir, "installer.properties")
            txt = '\n'.join([
                "INSTALLER_UI=SILENT",
                "USER_INSTALL_DIR=%s" % self.installdir,
                "MAKE_DEF_VER=true",
                "DOC_ROOT=UNDEFINED",
                "DOC_ROOT_TYPE=false",
                "DOC_ROOT_ESCAPED=UNDEFINED",
                "ABAQUSLM_LICENSE_FILE=@abaqusfea",
                "LICENSE_SERVER_TYPE=FLEXNET",
                "PRODUCT_NAME=Abaqus %s" % self.version,
                "TMPDIR=%s" % self.builddir,
                "INSTALL_MPI=1",
            ])
            write_file(self.replayfile, txt)

    def install_step(self):
        """Install ABAQUS using 'setup'."""
        if LooseVersion(self.version) >= LooseVersion('2016'):
            change_dir(os.path.join(self.cfg['start_dir'], '1'))
            qa = {
                "Enter selection (default: Install):": '',
                "Enter selection (default: Close):": '',
            }
            no_qa = [
                '___',
                r"\(\d+\s*[KM]B\)",
                r'\.\.\.$',
            ]

            # Match string for continuing on with the selected items
            nextstr = r"Enter selection \(default: Next\):\s*"

            # Allow for selection or deselection of components from lines of the form:
            #   5 [*] Tosca Fluid
            # This uses nextstr to make sure we only match the latest output in the Q&A process;
            # negative lookahead (?!___) is used to exclude ___...___ lines, to avoid matching across questions
            selectionstr = r"\s*(?P<nr>[-0-9]+) %%s %%s[ \w]*\n((?!%s)(?!___)[\S ]*\n)*%s$" % (nextstr, nextstr)

            # ensure question patterns are considered in order by using an OrderedDict instance,
            # rather than a regular dictionary (where there's no guarantee on key order in general)
            std_qa = OrderedDict()

            # Disable Extended Product Documentation because it has a troublesome Java dependency
            std_qa[selectionstr % (r"\[*\]", "Extended Product Documentation")] = "%(nr)s"
            installed_docs = False  # hard disabled, previous support was actually incomplete

            # enable all ABAQUS components
            std_qa[selectionstr % (r"\[ \]", "Abaqus.*")] = "%(nr)s"
            std_qa[selectionstr % (r"\[ \]", "Cosimulation Services")] = "%(nr)s"

            # enable 3DSFlow Solver (used to be called "Abaqus/CFD Solver")
            std_qa[selectionstr % (r"\[ \]", "3DSFlow Solver")] = "%(nr)s"

            # disable/enable fe-safe components
            if self.cfg['with_fe_safe']:
                std_qa[selectionstr % (r"\[ \]", ".*fe-safe")] = "%(nr)s"
            else:
                std_qa[selectionstr % (r"\[*\]", ".*fe-safe")] = "%(nr)s"

            # Disable/enable Tosca
            if self.cfg['with_tosca']:
                std_qa[selectionstr % (r"\[\ \]", "Tosca.*")] = "%(nr)s"
            else:
                std_qa[selectionstr % (r"\[\*\]", "Tosca.*")] = "%(nr)s"

            # disable CloudView
            std_qa[r"(?P<cloudview>[0-9]+) \[X\] Search using CloudView\nEnter selection:"] = '%(cloudview)s\n\n'
            # accept default port for documentation server
            std_qa[r"Check that the port is free.\nDefault \[[0-9]+\]:"] = '\n'

            # disable feedback by users
            std_qa[r"(?P<feedback>[0-9]+) \[X\] Allow users to send feedback.\nEnter selection:"] = '%(feedback)s\n\n'

            # disable reverse proxy
            std_qa[r"(?P<proxy>[0-9]+) \[X\] Use a reverse proxy.\nEnter selection:"] = '%(proxy)s\n\n'

            # Disable Isight
            std_qa[selectionstr % (r"\[\*\]", "Isight")] = "%(nr)s"
            # Disable Search using EXALEAD
            std_qa[r"\s*(?P<exalead>[0-9]+) \[X\] Search using EXALEAD\nEnter selection:"] = '%(exalead)s\n\n'

            # Directories
            cae_subdir = os.path.join(self.installdir, 'cae')
            sim_subdir = os.path.join(self.installdir, 'sim')
            std_qa[r"Default.*SIMULIA/EstProducts.*:"] = cae_subdir
            std_qa[r"SIMULIA[0-9]*doc.*:"] = os.path.join(self.installdir, 'doc')  # if docs are installed
            std_qa[r"SimulationServices.*:"] = sim_subdir
            std_qa[r"Choose the CODE installation directory.*:\n.*\n\n.*:"] = sim_subdir
            std_qa[r"SIMULIA/CAE.*:"] = cae_subdir
            std_qa[r"location of your Abaqus services \(solvers\).*(\n.*){8}:\s*"] = sim_subdir
            std_qa[r"Default.*SIMULIA/Commands\]:\s*"] = os.path.join(self.installdir, 'Commands')
            std_qa[r"Default.*SIMULIA/CAE/plugins.*:\s*"] = os.path.join(self.installdir, 'plugins')
            std_qa[r"Default.*SIMULIA/Isight.*:\s*"] = os.path.join(self.installdir, 'Isight')
            std_qa[r"Default.*SIMULIA/fe-safe/.*:"] = os.path.join(self.installdir, 'fe-safe')
            std_qa[r"Default.*SIMULIA/Tosca.*:"] = os.path.join(self.installdir, 'tosca')

            # paths to STAR-CCM+, FLUENT are requested when Tosca is also installed;
            # these do not strictly need to be specified at installation time, so we don't
            std_qa[r"STAR-CCM.*\n((?!___)[\S ]*\n)*\nDefault \[\]:"] = ''
            std_qa[r"FLUENT.*\n((?!___)[\S ]*\n)*\nDefault \[\]:"] = ''

            std_qa[r"location of your existing ANSA installation.*(\n.*){8}:"] = ''
            std_qa[r"FLUENT Path.*(\n.*){7}:"] = ''
            std_qa[r"working directory to be used by Tosca Fluid\s*(\n.*)*Default \[/usr/temp\]:\s*"] = '/tmp'

            # License server
            std_qa[r"License Server [0-9]+\s*(\n.*){3}:"] = 'abaqusfea'  # bypass value for license server
            std_qa[r"License Server . \(redundant\)\s*(\n.*){3}:"] = ''
            std_qa[r"License Server Configuration((?!___).*\n)*" + nextstr] = ''

            std_qa[r"Please choose an action:"] = '1'

            if LooseVersion(self.version) >= LooseVersion('2022') and installed_docs:
                java_root = get_software_root('Java')
                if java_root:
                    std_qa[r"Please enter .*Java Runtime Environment.* path.(\n.*)+Default \[\]:"] = java_root
                    std_qa[r"Please enter .*Java Runtime Environment.* path.(\n.*)+Default \[.+\]:"] = ''
                else:
                    raise EasyBuildError("Java is required for ABAQUS docs versions >= 2022, but it is missing")

            # Continue
            std_qa[nextstr] = ''

            run_cmd_qa('./StartTUI.sh', qa, no_qa=no_qa, std_qa=std_qa, log_all=True, simple=True, maxhits=1000)
        else:
            change_dir(self.builddir)
            if self.cfg['install_cmd'] is None:
                self.cfg['install_cmd'] = "%s/%s-%s/setup" % (self.builddir, self.name, self.version.split('-')[0])
                self.cfg['install_cmd'] += " -replay %s" % self.replayfile
                if LooseVersion(self.version) < LooseVersion("6.13"):
                    self.cfg['install_cmd'] += " -nosystemcheck"
            super(EB_ABAQUS, self).install_step()

        if LooseVersion(self.version) >= LooseVersion('2016'):
            # also install hot fixes (if any)
            hotfixes = [src for src in self.src if 'CFA' in src['name']]
            if hotfixes:

                # first install Part_3DEXP_SimulationServices hotfix(es), if any
                hotfixes_3dexp = [src for src in self.src if 'CFA' in src['name'] and '3DEXP' in src['name']]
                if hotfixes_3dexp:
                    hotfix_dir = os.path.join(self.builddir, 'Part_3DEXP_SimulationServices.Linux64', '1', 'Software')
                    change_dir(hotfix_dir)

                    # SIMULIA_ComputeServices part
                    subdirs = glob.glob('HF_SIMULIA_ComputeServices.HF*.Linux64')
                    if len(subdirs) == 1:
                        subdir = subdirs[0]
                    else:
                        raise EasyBuildError("Failed to find expected subdir for hotfix: %s", subdirs)

                    cwd = change_dir(os.path.join(subdir, '1'))
                    std_qa = OrderedDict()
                    std_qa[r"Enter selection \(default: Next\):"] = ''
                    std_qa["Choose the .*installation directory.*\n.*\n\n.*:"] = os.path.join(self.installdir, 'sim')
                    std_qa[r"Enter selection \(default: Install\):"] = ''
                    run_cmd_qa('./StartTUI.sh', {}, std_qa=std_qa, log_all=True, simple=True, maxhits=100)

                    # F_CAASIMULIAComputeServicesBuildTime part
                    change_dir(cwd)
                    subdirs = glob.glob('HF_CAASIMULIAComputeServicesBuildTime.HF*.Linux64')
                    if len(subdirs) == 1:
                        subdir = subdirs[0]
                    else:
                        raise EasyBuildError("Failed to find expected subdir for hotfix: %s", subdirs)

                    cwd = change_dir(os.path.join(cwd, subdir, '1'))
                    run_cmd_qa('./StartTUI.sh', {}, std_qa=std_qa, log_all=True, simple=True, maxhits=100)
                    change_dir(cwd)

                # next install Part_SIMULIA_Abaqus_CAE hotfix (ABAQUS versions <= 2020)
                hotfix_dir = os.path.join(self.builddir, 'Part_SIMULIA_Abaqus_CAE.Linux64', '1', 'Software')
                if os.path.exists(hotfix_dir):
                    change_dir(hotfix_dir)

                    subdirs = glob.glob('SIMULIA_Abaqus_CAE.HF*.Linux64')
                    if len(subdirs) == 1:
                        subdir = subdirs[0]
                    else:
                        raise EasyBuildError("Failed to find expected subdir for hotfix: %s", subdirs)

                    cwd = change_dir(os.path.join(subdir, '1'))
                    std_qa = OrderedDict()
                    std_qa[r"Enter selection \(default: Next\):"] = ''
                    std_qa["Choose the .*installation directory.*\n.*\n\n.*:"] = os.path.join(self.installdir, 'cae')
                    std_qa[r"Enter selection \(default: Install\):"] = ''
                    std_qa[r"\[1\] Continue\n(?:.|\n)*Please choose an action:"] = '1'
                    std_qa[r"\[2\] Continue\n(?:.|\n)*Please choose an action:"] = '2'
                    no_qa = [r"Please be patient;  it will take a few minutes to complete\.\n(\.)*"]
                    run_cmd_qa('./StartTUI.sh', {}, no_qa=no_qa, std_qa=std_qa, log_all=True, simple=True, maxhits=100)
                    change_dir(cwd)

                # install SIMULIA Established Products hotfix (ABAQUS versions > 2020)
                hotfixes_estprd = [src for src in self.src if 'CFA' in src['name'] and 'EstPrd' in src['name']]
                if hotfixes_estprd:
                    hotfix_dir = os.path.join(self.builddir, 'Part_SIMULIA_EstPrd.Linux64', '1', 'Software')
                    change_dir(hotfix_dir)

                    subdirs = glob.glob('SIMULIA_EstPrd.HF*.Linux64')
                    if len(subdirs) == 1:
                        subdir = subdirs[0]
                    else:
                        raise EasyBuildError("Failed to find expected subdir for hotfix: %s", subdirs)

                    cwd = change_dir(os.path.join(subdir, '1'))
                    no_qa = [
                        '___',
                        '...',
                        r'\(\d+[KM]B\)',
                    ]
                    std_qa = OrderedDict()
                    std_qa[r"Enter selection \(default: Next\):"] = ''
                    std_qa["Choose the .*installation directory.*\n.*\n\n.*:"] = os.path.join(self.installdir, 'cae')
                    std_qa[r"Enter selection \(default: Install\):"] = ''
                    std_qa[r"The Abaqus commands directory.*:\n.*\n+Actions:\n.*\n_+\n\nPlease.*:"] = '1'
                    std_qa[r"Enter selection \(default: Close\):"] = ''

                    run_cmd_qa('./StartTUI.sh', {}, no_qa=no_qa, std_qa=std_qa, log_all=True, simple=True, maxhits=100)
                    change_dir(cwd)

        # create 'abaqus' symlink for main command, which is not there anymore starting with ABAQUS 2022
        if LooseVersion(self.version) >= LooseVersion('2022'):
            commands_dir = os.path.join(self.installdir, 'Commands')
            abaqus_cmd = os.path.join(commands_dir, 'abaqus')
            if os.path.exists(abaqus_cmd):
                self.log.info("Main 'abaqus' command already found at %s, no need to create symbolic link", abaqus_cmd)
            else:
                abq_ver_cmd = os.path.join(commands_dir, 'abq%s' % self.version)
                self.log.info("Creating symbolic link 'abaqus' for main command %s", abq_ver_cmd)
                if os.path.exists(abq_ver_cmd):
                    cwd = change_dir(commands_dir)
                    symlink(os.path.basename(abq_ver_cmd), os.path.basename(abaqus_cmd))
                    change_dir(cwd)
                else:
                    raise EasyBuildError("Path to main command %s does not exist!", abq_ver_cmd)

    def sanity_check_step(self):
        """Custom sanity check for ABAQUS."""
        custom_paths = {
            'files': [os.path.join('Commands', 'abaqus')],
            'dirs': [],
        }
        custom_commands = []

        if LooseVersion(self.version) >= LooseVersion('2016'):
            custom_paths['dirs'].extend(['cae', 'Commands'])
            if LooseVersion(self.version) < LooseVersion('2020'):
                custom_paths['dirs'].extend(['sim'])
            # 'all' also check license server, but lmstat is usually not available
            custom_commands.append("abaqus information=system")
        else:
            verparts = self.version.split('-')[0].split('.')
            custom_paths['dirs'].append('%s-%s' % ('.'.join(verparts[0:2]), verparts[2]))
            custom_commands.append("abaqus information=all")

        if LooseVersion(self.version) >= LooseVersion('2020'):
            custom_paths['files'].append(os.path.join('cae', 'linux_a64', 'code', 'bin', 'abaqusstd'))
            if self.cfg['with_fe_safe']:
                custom_paths['files'].append(os.path.join('cae', 'linux_a64', 'code', 'bin', 'fe-safe'))

        if self.cfg['with_tosca']:
            custom_commands.append("ToscaPython.sh --help")

        super(EB_ABAQUS, self).sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)

    def make_module_req_guess(self):
        """Update $PATH guesses for ABAQUS."""
        guesses = super(EB_ABAQUS, self).make_module_req_guess()

        path_subdirs = ['Commands']
        if self.cfg['with_tosca']:
            path_subdirs.append(os.path.join('cae', 'linux_a64', 'code', 'command'))

        guesses.update({
            'PATH': path_subdirs,
        })

        return guesses



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing binutils, implemented as an easyblock

@author: Kenneth Hoste (HPC-UGent)
"""
import glob
import os
import re
from distutils.version import LooseVersion

import easybuild.tools.environment as env
from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import apply_regex_substitutions, copy_file
from easybuild.tools.modules import get_software_libdir, get_software_root
from easybuild.tools.run import run_cmd
from easybuild.tools.systemtools import RISCV, get_cpu_family, get_shared_lib_ext
from easybuild.tools.utilities import nub


class EB_binutils(ConfigureMake):
    """Support for building/installing binutils."""

    @staticmethod
    def extra_options(extra_vars=None):
        """Extra easyconfig parameters specific to the binutils easyblock."""
        extra_vars = ConfigureMake.extra_options(extra_vars=extra_vars)
        extra_vars.update({
            'install_libiberty': [True, "Also install libiberty (implies building with -fPIC)", CUSTOM],
            'use_debuginfod': [False, "Build with debuginfod (used from system)", CUSTOM],
        })
        return extra_vars

    def __init__(self, *args, **kwargs):
        """Easyblock constructor"""
        super(EB_binutils, self).__init__(*args, **kwargs)

        # ld.gold linker is not supported on RISC-V
        self.use_gold = get_cpu_family() != RISCV

    def determine_used_library_paths(self):
        """Check which paths are used to search for libraries"""

        # determine C compiler command: use $CC, fall back to 'gcc' (when using system toolchain)
        compiler_cmd = os.environ.get('CC', 'gcc')

        # determine library search paths for GCC
        stdout, ec = run_cmd('LC_ALL=C "%s" -print-search-dirs' % compiler_cmd, simple=False, log_all=True)
        if ec:
            raise EasyBuildError("Failed to determine library search dirs from compiler %s", compiler_cmd)

        m = re.search('^libraries: *=(.*)$', stdout, re.M)
        paths = nub(os.path.abspath(p) for p in m.group(1).split(os.pathsep))
        self.log.debug('Unique library search paths from compiler %s: %s', compiler_cmd, paths)

        # Filter out all paths that do not exist
        paths = [p for p in paths if os.path.exists(p)]
        self.log.debug("Existing library search paths: %s", ', '.join(paths))

        result = []
        for path in paths:
            if any(os.path.samefile(path, p) for p in result):
                self.log.debug("Skipping symlink to existing path: %s", path)
            elif not glob.glob(os.path.join(path, '*.so*')):
                self.log.debug("Skipping path with no shared libraries: %s", path)
            else:
                result.append(path)

        self.log.info("Determined library search paths: %s", ', '.join(result))
        return result

    def configure_step(self):
        """Custom configuration procedure for binutils: statically link to zlib, configure options."""

        if self.toolchain.is_system_toolchain():
            # determine list of 'lib' directories to use rpath for;
            # this should 'harden' the resulting binutils to bootstrap GCC
            # (no trouble when other libstdc++ is build etc)
            lib_paths = self.determine_used_library_paths()

            # The installed lib dir must come first though to avoid taking system libs over installed ones, see:
            # https://github.com/easybuilders/easybuild-easyconfigs/issues/10056
            # To get literal $ORIGIN through Make we need to escape it by doubling $$, else it's a variable to Make;
            # We need to include both 'lib' and 'lib64' here, to ensure this works as intended
            # across different operating systems,
            # see https://github.com/easybuilders/easybuild-easyconfigs/issues/11976
            lib_paths = [r'$$ORIGIN/../lib', r'$$ORIGIN/../lib64'] + lib_paths
            # Mind the single quotes
            libs = ["-Wl,-rpath='%s'" % x for x in lib_paths]
        else:
            libs = []

        binutilsroot = get_software_root('binutils')
        if binutilsroot:
            # Remove LDFLAGS that start with '-L' + binutilsroot, since we don't
            # want to link libraries from binutils compiled with the system toolchain
            # into binutils binaries compiled with a compiler toolchain.
            ldflags = os.getenv('LDFLAGS').split(' ')
            ldflags = [p for p in ldflags if not p.startswith('-L' + binutilsroot)]
            env.setvar('LDFLAGS', ' '.join(ldflags))

        # configure using `--with-system-zlib` if zlib is a (build) dependency
        zlibroot = get_software_root('zlib')
        if zlibroot:
            self.cfg.update('configopts', '--with-system-zlib')

            # statically link to zlib only if it is a build dependency
            # see https://github.com/easybuilders/easybuild-easyblocks/issues/1350
            build_deps = self.cfg.dependencies(build_only=True)
            if any(dep['name'] == 'zlib' for dep in build_deps):
                libz_path = os.path.join(zlibroot, get_software_libdir('zlib'), 'libz.a')

                # for recent binutils versions, we need to override ZLIB in Makefile.in of components
                if LooseVersion(self.version) >= LooseVersion('2.26'):
                    regex_subs = [
                        (r"^(ZLIB\s*=\s*).*$", r"\1%s" % libz_path),
                        (r"^(ZLIBINC\s*=\s*).*$", r"\1-I%s" % os.path.join(zlibroot, 'include')),
                    ]
                    for makefile in glob.glob(os.path.join(self.cfg['start_dir'], '*', 'Makefile.in')):
                        apply_regex_substitutions(makefile, regex_subs)

                # for older versions, injecting the path to the static libz library into $LIBS works
                else:
                    libs.append(libz_path)

        env.setvar('LIBS', ' '.join(libs))

        # explicitly configure binutils to use / as sysroot
        # this is required to ensure the binutils installation works correctly with a (system)
        # GCC compiler that was explicitly configured with --with-sysroot=/;
        # we should *not* use the value of the EasyBuild --sysroot configuration option here,
        # since that leads to weird errors where the sysroot path is duplicated, like:
        #   /bin/ld.gold: error: cannot open /<sysroot>/<sysroot>/lib64/libc.so.6: No such file or directory
        # (see also https://gcc.gnu.org/legacy-ml/gcc-help/2006-08/msg00212.html)
        self.cfg.update('configopts', '--with-sysroot=/')

        # build both static and shared libraries for recent binutils versions (default is only static)
        if LooseVersion(self.version) > LooseVersion('2.24'):
            self.cfg.update('configopts', "--enable-shared --enable-static")

        # enable gold linker with plugin support, use ld as default linker (for recent versions of binutils)
        if LooseVersion(self.version) > LooseVersion('2.24'):
            self.cfg.update('configopts', "--enable-plugins --enable-ld=default")
            if self.use_gold:
                self.cfg.update('configopts', '--enable-gold')

        if LooseVersion(self.version) >= LooseVersion('2.34'):
            if self.cfg['use_debuginfod']:
                self.cfg.update('configopts', '--with-debuginfod')
            else:
                self.cfg.update('configopts', '--without-debuginfod')

        # complete configuration with configure_method of parent
        super(EB_binutils, self).configure_step()

        if self.cfg['install_libiberty']:
            cflags = os.getenv('CFLAGS')
            if cflags:
                self.cfg.update('buildopts', 'CFLAGS="$CFLAGS -fPIC"')
            else:
                # if $CFLAGS is not defined, make sure we retain "-g -O2",
                # since not specifying any optimization level implies -O0...
                self.cfg.update('buildopts', 'CFLAGS="-g -O2 -fPIC"')

    def install_step(self):
        """Install using 'make install', also install libiberty if desired."""
        super(EB_binutils, self).install_step()

        # only install libiberty files if if they're not there yet;
        # libiberty.a is installed by default for old binutils versions
        if self.cfg['install_libiberty']:
            if not os.path.exists(os.path.join(self.installdir, 'include', 'libiberty.h')):
                copy_file(os.path.join(self.cfg['start_dir'], 'include', 'libiberty.h'),
                          os.path.join(self.installdir, 'include', 'libiberty.h'))

            if not glob.glob(os.path.join(self.installdir, 'lib*', 'libiberty.a')):
                # Be a bit careful about where we install into
                libdir = os.path.join(self.installdir, 'lib')
                # At this point the lib directory should exist and be populated, if not try the other option
                if not (os.path.exists(libdir) and os.path.isdir(libdir) and os.listdir(libdir)):
                    libdir = os.path.join(self.installdir, 'lib64')

                # Make sure the target exists (it should, otherwise our sanity check will fail)
                if os.path.exists(libdir) and os.path.isdir(libdir) and os.listdir(libdir):
                    copy_file(os.path.join(self.cfg['start_dir'], 'libiberty', 'libiberty.a'),
                              os.path.join(libdir, 'libiberty.a'))
                else:
                    raise EasyBuildError("Target installation directory %s for libiberty.a is non-existent or empty",
                                         libdir)

            if not os.path.exists(os.path.join(self.installdir, 'info', 'libiberty.texi')):
                copy_file(os.path.join(self.cfg['start_dir'], 'libiberty', 'libiberty.texi'),
                          os.path.join(self.installdir, 'info', 'libiberty.texi'))

    def sanity_check_step(self):
        """Custom sanity check for binutils."""

        binaries = ['addr2line', 'ar', 'as', 'c++filt', 'elfedit', 'gprof', 'ld', 'ld.bfd', 'nm',
                    'objcopy', 'objdump', 'ranlib', 'readelf', 'size', 'strings', 'strip']

        headers = ['ansidecl.h', 'bfd.h', 'bfdlink.h', 'dis-asm.h', 'symcat.h']
        libs = ['bfd', 'opcodes']

        lib_exts = ['a']
        shlib_ext = get_shared_lib_ext()

        if LooseVersion(self.version) > LooseVersion('2.24'):
            lib_exts.append(shlib_ext)
            if self.use_gold:
                binaries.append('ld.gold')

        bin_paths = [os.path.join('bin', b) for b in binaries]
        inc_paths = [os.path.join('include', h) for h in headers]

        libs_fn = ['lib%s.%s' % (lib, ext) for lib in libs for ext in lib_exts]
        lib_paths = [(os.path.join('lib', lib_fn), os.path.join('lib64', lib_fn)) for lib_fn in libs_fn]

        custom_paths = {
            'files': bin_paths + inc_paths + lib_paths,
            'dirs': [],
        }

        if self.cfg['install_libiberty']:
            custom_paths['files'].extend([
                (os.path.join('lib', 'libiberty.a'), os.path.join('lib64', 'libiberty.a')),
                os.path.join('include', 'libiberty.h'),
            ])

        # All binaries support --version, check that they can be run
        custom_commands = ['%s --version' % b for b in binaries]

        # if zlib is listed as a build dependency, it should have been linked in statically
        build_deps = self.cfg.dependencies(build_only=True)
        if any(dep['name'] == 'zlib' for dep in build_deps):
            for binary in binaries:
                bin_path = os.path.join(self.installdir, 'bin', binary)
                out, _ = run_cmd("file %s" % bin_path, simple=False)
                if re.search(r'statically linked', out):
                    # binary is fully statically linked, so no chance for dynamically linked libz
                    continue

                # check whether libz is linked dynamically, it shouldn't be
                out, _ = run_cmd("ldd %s" % bin_path, simple=False)
                if re.search(r'libz\.%s' % shlib_ext, out):
                    raise EasyBuildError("zlib is not statically linked in %s: %s", bin_path, out)

        super(EB_binutils, self).sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing MRtrix, implemented as an easyblock
"""
import glob
import os
from distutils.version import LooseVersion

import easybuild.tools.environment as env
from easybuild.framework.easyblock import EasyBlock
from easybuild.tools.filetools import copy, symlink
from easybuild.tools.run import run_cmd
from easybuild.tools.systemtools import get_shared_lib_ext


class EB_MRtrix(EasyBlock):
    """Support for building/installing MRtrix."""

    def __init__(self, *args, **kwargs):
        """Initialize easyblock, enable build-in-installdir based on version."""
        super(EB_MRtrix, self).__init__(*args, **kwargs)

        if LooseVersion(self.version) >= LooseVersion('0.3') and LooseVersion(self.version) < LooseVersion('0.3.14'):
            self.build_in_installdir = True
            self.log.debug("Enabled build-in-installdir for version %s", self.version)

    def extract_step(self):
        """Extract MRtrix sources."""
        # strip off 'mrtrix*' part to avoid having everything in a 'mrtrix*' subdirectory
        if LooseVersion(self.version) >= LooseVersion('0.3'):
            self.cfg.update('unpack_options', '--strip-components=1')

        super(EB_MRtrix, self).extract_step()

    def configure_step(self):
        """No configuration step for MRtrix."""
        if LooseVersion(self.version) >= LooseVersion('0.3'):
            if LooseVersion(self.version) < LooseVersion('0.3.13'):
                env.setvar('LD', "%s LDFLAGS OBJECTS -o EXECUTABLE" % os.getenv('CXX'))
                env.setvar('LDLIB', "%s -shared LDLIB_FLAGS OBJECTS -o LIB" % os.getenv('CXX'))

            env.setvar('QMAKE_CXX', os.getenv('CXX'))
            cmd = "python configure -verbose"

            run_cmd(cmd, log_all=True, simple=True, log_ok=True)

    def build_step(self):
        """Custom build procedure for MRtrix."""
        parallel = self.cfg['parallel']
        env.setvar('NUMBER_OF_PROCESSORS', str(parallel))

        cmd = "python build -verbose"
        run_cmd(cmd, log_all=True, simple=True, log_ok=True)

    def install_step(self):
        """Custom install procedure for MRtrix."""
        if LooseVersion(self.version) < LooseVersion('0.3'):
            cmd = "python build -verbose install=%s linkto=" % self.installdir
            run_cmd(cmd, log_all=True, simple=True, log_ok=True)

        elif LooseVersion(self.version) >= LooseVersion('3.0'):
            copy(os.path.join(self.builddir, 'bin'), self.installdir)
            copy(os.path.join(self.builddir, 'lib'), self.installdir)

        elif LooseVersion(self.version) >= LooseVersion('0.3.14'):
            copy(glob.glob(os.path.join(self.builddir, 'release', '*')), self.installdir)
            copy(os.path.join(self.builddir, 'scripts'), self.installdir)
            # some scripts expect 'release/bin' to be there, so we put a symlink in place
            symlink(self.installdir, os.path.join(self.installdir, 'release'))

    def make_module_req_guess(self):
        """
        Return list of subdirectories to consider to update environment variables;
        also consider 'scripts' subdirectory for $PATH
        """
        guesses = super(EB_MRtrix, self).make_module_req_guess()

        guesses['PATH'].append('scripts')

        if LooseVersion(self.version) >= LooseVersion('3.0'):
            guesses.setdefault('PYTHONPATH', []).append('lib')

        return guesses

    def sanity_check_step(self):
        """Custom sanity check for MRtrix."""
        shlib_ext = get_shared_lib_ext()

        if LooseVersion(self.version) >= LooseVersion('0.3'):
            libso = 'libmrtrix.%s' % shlib_ext
        else:
            libso = 'libmrtrix-%s.%s' % ('_'.join(self.version.split('.')), shlib_ext)
        custom_paths = {
            'files': [os.path.join('lib', libso)],
            'dirs': ['bin'],
        }

        custom_commands = []
        if LooseVersion(self.version) >= LooseVersion('3.0'):
            custom_commands.append("python -c 'import mrtrix3'")

        super(EB_MRtrix, self).sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for bzip2, implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
@author: Jens Timmerman (Ghent University)
@author: Pablo Escobar (sciCORE, UniBas)
"""
import glob
import os
import shutil

from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.run import run_cmd
from easybuild.tools.systemtools import get_shared_lib_ext


class EB_bzip2(ConfigureMake):
    """Support for building and installing bzip2."""

    @staticmethod
    def extra_options():
        """Add extra config options specific to bzip2."""
        extra_vars = {
            'with_shared_libs': [True, "Also build shared libraries", CUSTOM],
        }
        return ConfigureMake.extra_options(extra_vars=extra_vars)

    # no configure script
    def configure_step(self):
        """Set extra options for 'make' command (CC, CFLAGS)."""

        if 'CC=' not in self.cfg['buildopts']:
            self.cfg.update('buildopts', 'CC="%s"' % os.getenv('CC'))
        if 'CFLAGS=' not in self.cfg['buildopts']:
            self.cfg.update('buildopts', "CFLAGS='-Wall -Winline %s -g $(BIGFILES)'" % os.getenv('CFLAGS'))

    def install_step(self):
        """Install in non-standard path by passing PREFIX variable to make install."""

        self.cfg.update('installopts', "PREFIX=%s" % self.installdir)
        super(EB_bzip2, self).install_step()

        # also build & install shared libraries, if desired
        if self.cfg['with_shared_libs']:

            cmd = "%s make -f Makefile-libbz2_so %s" % (self.cfg['prebuildopts'], self.cfg['buildopts'])
            run_cmd(cmd, log_all=True, simple=True)

            # copy shared libraries to <install dir>/lib
            shlib_ext = get_shared_lib_ext()
            libdir = os.path.join(self.installdir, 'lib')
            try:
                for lib in glob.glob('libbz2.%s.*' % shlib_ext):
                    # only way to copy a symlink is to check for it,
                    # cfr. http://stackoverflow.com/questions/4847615/copying-a-symbolic-link-in-python
                    if os.path.islink(lib):
                        os.symlink(os.readlink(lib), os.path.join(libdir, lib))
                    else:
                        shutil.copy2(lib, libdir)
            except OSError as err:
                raise EasyBuildError("Copying shared libraries to installation dir %s failed: %s", libdir, err)

            # create symlink libbz2.so >> libbz2.so.1.0.6
            try:
                cwd = os.getcwd()
                os.chdir(libdir)
                os.symlink('libbz2.%s.%s' % (shlib_ext, self.version), 'libbz2.%s' % shlib_ext)
                os.chdir(cwd)
            except OSError as err:
                raise EasyBuildError("Creating symlink for libbz2.so failed: %s", err)

    def sanity_check_step(self):
        """Custom sanity check for bzip2."""
        libs = ['lib/libbz2.a']
        if self.cfg['with_shared_libs']:
            shlib_ext = get_shared_lib_ext()
            libs.extend(['lib/libbz2.%s.%s' % (shlib_ext, self.version), 'lib/libbz2.%s' % shlib_ext])

        custom_paths = {
            'files': ['bin/b%s' % x for x in ['unzip2', 'zcat', 'zdiff', 'zgrep', 'zip2', 'zip2recover', 'zmore']] +
            ['include/bzlib.h'] + libs,
            'dirs': [],
        }
        super(EB_bzip2, self).sanity_check_step(custom_paths=custom_paths)



##
# This file is an EasyBuild reciPY as per https://github.com/easybuilders/easybuild
#
# Copyright:: Copyright 2012-2023 Uni.Lu/LCSB, NTUA
# Authors::   Cedric Laczny <cedric.laczny@uni.lu>, Kenneth Hoste
# Authors::   George Tsouloupas <g.tsouloupas@cyi.ac.cy>, Fotis Georgatos <fotis@cern.ch>
# License::   MIT/GPL
# $Id$
#
# This work implements a part of the HPCBIOS project and is a component of the policy:
# http://hpcbios.readthedocs.org/en/latest/HPCBIOS_2012-94.html
##
"""
EasyBuild support for building and installing BWA, implemented as an easyblock

@author: Cedric Laczny (Uni.Lu)
@author: Fotis Georgatos (Uni.Lu)
@author: Kenneth Hoste (Ghent University)
@author: George Tsouloupas <g.tsouloupas@cyi.ac.cy>
"""
import os
import glob
from distutils.version import LooseVersion

from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import copy_file, mkdir


class EB_BWA(ConfigureMake):
    """
    Support for building BWA
    """

    def __init__(self, *args, **kwargs):
        """Add extra config options specific to BWA."""
        super(EB_BWA, self).__init__(*args, **kwargs)

        self.files = ['bwa', 'qualfa2fq.pl', 'xa2multi.pl']
        if LooseVersion(self.version) < LooseVersion('0.7.0'):
            # solid2fastq was dropped in recent versions because the same functionality
            # is covered by other tools already
            # cfr. http://osdir.com/ml/general/2010-10/msg26205.html
            self.files.append('solid2fastq.pl')
        self.includes = []
        self.libs = ['libbwa.a']

    def configure_step(self):
        """
        Empty function as BWA comes with _no_ configure script
        """
        pass

    def build_step(self):
        """Custom build procedure: pass down compiler command and options as arguments to 'make'."""

        for env_var in ('CC', 'CFLAGS'):
            if env_var + '=' not in self.cfg['buildopts']:
                self.cfg.update('buildopts', env_var + '="$' + env_var + '"')

        super(EB_BWA, self).build_step()

    def install_step(self):
        """
        Install by copying files to install dir
        """
        srcdir = self.cfg['start_dir']
        # copy binaries
        bindir = os.path.join(self.installdir, 'bin')
        mkdir(bindir)
        for filename in self.files:
            srcfile = os.path.join(srcdir, filename)
            copy_file(srcfile, bindir)

        # copy include files
        includes = glob.glob(os.path.join(srcdir, '*.h'))
        self.includes = [os.path.basename(include) for include in includes]
        incdir = os.path.join(self.installdir, 'include', 'bwa')
        if not self.includes:
            raise EasyBuildError("Unable to find header files")

        mkdir(incdir, parents=True)
        for filename in self.includes:
            srcfile = os.path.join(srcdir, filename)
            copy_file(srcfile, incdir)

        # copy libraries
        libdir = os.path.join(self.installdir, 'lib')
        mkdir(libdir)
        for filename in self.libs:
            srcfile = os.path.join(srcdir, filename)
            copy_file(srcfile, libdir)

        manfile = os.path.join(srcdir, 'bwa.1')
        manman1dir = os.path.join(self.installdir, 'man', 'man1')
        mkdir(manman1dir, parents=True)
        copy_file(manfile, manman1dir)

    def sanity_check_step(self):
        """Custom sanity check for BWA."""

        bins = [os.path.join('bin', x) for x in self.files]
        incs = [os.path.join('include', 'bwa', x) for x in self.includes]
        libs = [os.path.join('lib', 'libbwa.a')]

        custom_paths = {
            'files': bins + incs + libs,
            'dirs': []
        }

        # 'bwa' command doesn't have a --help option, but it does print help-like information to stderr
        # when run without arguments (and exits with exit code 1)
        custom_commands = ["bwa 2>&1 | grep 'index sequences in the FASTA format'"]

        super(EB_BWA, self).sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for SLEPc, implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
"""

import os
import re
from distutils.version import LooseVersion

import easybuild.tools.environment as env
from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.framework.easyconfig import BUILD, CUSTOM
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.modules import get_software_root
from easybuild.tools.run import run_cmd


class EB_SLEPc(ConfigureMake):
    """Support for building and installing SLEPc"""

    @staticmethod
    def extra_options():
        """Add extra config options specific to SLEPc."""
        extra_vars = {
            'runtest': ['test', "Make target to test build", BUILD],
            'petsc_arch': [None, "PETSc architecture to use (value for $PETSC_ARCH)", CUSTOM],
            'sourceinstall': [False, "Indicates whether a source installation should be performed", CUSTOM],
        }
        return ConfigureMake.extra_options(extra_vars)

    def __init__(self, *args, **kwargs):
        """Initialize SLEPc custom variables."""
        super(EB_SLEPc, self).__init__(*args, **kwargs)

        self.slepc_subdir = ''

        if self.cfg['sourceinstall']:
            self.build_in_installdir = True

    def configure_step(self):
        """Configure SLEPc by setting configure options and running configure script."""

        # check PETSc dependency
        if not get_software_root("PETSc"):
            raise EasyBuildError("PETSc module not loaded?")

        # set SLEPC_DIR environment variable
        env.setvar('SLEPC_DIR', self.cfg['start_dir'])
        self.log.debug('SLEPC_DIR: %s' % os.getenv('SLEPC_DIR'))

        # optional dependencies
        dep_filter = [d['name'] for d in self.cfg.builddependencies()] + ['PETSc', 'Python']
        deps = [dep['name'] for dep in self.cfg.dependencies() if dep['name'] not in dep_filter]
        for dep in deps:
            deproot = get_software_root(dep)
            if deproot:
                withdep = "--with-%s" % dep.lower()
                self.cfg.update('configopts', '%s=1 %s-dir=%s' % (withdep, withdep, deproot))

        if self.cfg['sourceinstall']:
            # run configure without --prefix (required)
            cmd = "%s ./configure %s" % (self.cfg['preconfigopts'], self.cfg['configopts'])
            (out, _) = run_cmd(cmd, log_all=True, simple=False)
        else:
            # regular './configure --prefix=X' for non-source install

            # make sure old install dir is removed first
            self.make_installdir(dontcreate=True)

            out = super(EB_SLEPc, self).configure_step()

        # check for errors in configure
        error_regexp = re.compile("ERROR")
        if error_regexp.search(out):
            raise EasyBuildError("Error(s) detected in configure output!")

        # define $PETSC_ARCH
        petsc_arch = self.cfg['petsc_arch']
        if self.cfg['petsc_arch'] is None:
            petsc_arch = 'arch-installed-petsc'

        env.setvar('PETSC_ARCH', petsc_arch)

        if self.cfg['sourceinstall']:
            self.slepc_subdir = os.path.join('%s-%s' % (self.name.lower(), self.version), petsc_arch)

        # SLEPc > 3.5, make does not accept -j
        if LooseVersion(self.version) >= LooseVersion("3.5"):
            self.cfg['parallel'] = None

    def make_module_req_guess(self):
        """Specify correct LD_LIBRARY_PATH and CPATH for SLEPc installation."""
        guesses = super(EB_SLEPc, self).make_module_req_guess()

        guesses.update({
            'CPATH': [os.path.join(self.slepc_subdir, "include")],
            'LD_LIBRARY_PATH': [os.path.join(self.slepc_subdir, "lib")],
        })

        return guesses

    def make_module_extra(self):
        """Set SLEPc specific environment variables (SLEPC_DIR)."""
        txt = super(EB_SLEPc, self).make_module_extra()

        if self.cfg['sourceinstall']:
            subdir = '%s-%s' % (self.name.lower(), self.version)
            txt += self.module_generator.set_environment('SLEPC_DIR', os.path.join(self.installdir, subdir))

        else:
            txt += self.module_generator.set_environment('SLEPC_DIR', self.installdir)

        return txt

    def sanity_check_step(self):
        """Custom sanity check for SLEPc"""
        custom_paths = {
            'files': [],
            'dirs': [os.path.join(self.slepc_subdir, x) for x in ['include', 'lib']],
        }
        if LooseVersion(self.version) < LooseVersion('3.6'):
            custom_paths['dirs'].append('conf')
        else:
            custom_paths['dirs'].append(os.path.join('lib', 'slepc', 'conf'))
        super(EB_SLEPc, self).sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2013-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing GROMACS, implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
@author: Ward Poelmans (Ghent University)
@author: Benjamin Roberts (The University of Auckland)
@author: Luca Marsella (CSCS)
@author: Guilherme Peretti-Pezzi (CSCS)
@author: Oliver Stueker (Compute Canada/ACENET)
@author: Davide Vanzo (Vanderbilt University)
"""
import glob
import os
import re
import shutil
from distutils.version import LooseVersion

import easybuild.tools.environment as env
import easybuild.tools.toolchain as toolchain
from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.easyblocks.generic.cmakemake import CMakeMake
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError, print_warning
from easybuild.tools.config import build_option
from easybuild.tools.filetools import copy_dir, find_backup_name_candidate, remove_dir, which
from easybuild.tools.modules import get_software_libdir, get_software_root, get_software_version
from easybuild.tools.run import run_cmd
from easybuild.tools.toolchain.compiler import OPTARCH_GENERIC
from easybuild.tools.systemtools import X86_64, get_cpu_architecture, get_shared_lib_ext, get_cpu_features
from easybuild.tools.version import VERBOSE_VERSION as EASYBUILD_VERSION


class EB_GROMACS(CMakeMake):
    """Support for building/installing GROMACS."""

    @staticmethod
    def extra_options():
        extra_vars = CMakeMake.extra_options()
        extra_vars.update({
            'double_precision': [None, "Build with double precision enabled (-DGMX_DOUBLE=ON), " +
                                 "default is to build double precision unless CUDA is enabled", CUSTOM],
            'mpisuffix': ['_mpi', "Suffix to append to MPI-enabled executables (only for GROMACS < 4.6)", CUSTOM],
            'mpiexec': ['mpirun', "MPI executable to use when running tests", CUSTOM],
            'mpiexec_numproc_flag': ['-np', "Flag to introduce the number of MPI tasks when running tests", CUSTOM],
            'mpi_numprocs': [0, "Number of MPI tasks to use when running tests", CUSTOM],
            'ignore_plumed_version_check': [False, "Ignore the version compatibility check for PLUMED", CUSTOM],
            'plumed': [None, "Try to apply PLUMED patches. None (default) is auto-detect. " +
                       "True or False forces behaviour.", CUSTOM],
        })
        extra_vars['separate_build_dir'][0] = True
        return extra_vars

    def __init__(self, *args, **kwargs):
        """Initialize GROMACS-specific variables."""
        super(EB_GROMACS, self).__init__(*args, **kwargs)
        self.lib_subdir = ''
        self.pre_env = ''
        self.cfg['build_shared_libs'] = self.cfg.get('build_shared_libs', False)
        if LooseVersion(self.version) >= LooseVersion('2019'):
            # Building the gmxapi interface requires shared libraries
            self.cfg['build_shared_libs'] = True
        if self.cfg['build_shared_libs']:
            self.libext = get_shared_lib_ext()
        else:
            self.libext = 'a'

    def get_gromacs_arch(self):
        """Determine value of GMX_SIMD CMake flag based on optarch string.

        Refs:
        [0] http://manual.gromacs.org/documentation/2016.3/install-guide/index.html#typical-installation
        [1] http://manual.gromacs.org/documentation/2016.3/install-guide/index.html#simd-support
        [2] http://www.gromacs.org/Documentation/Acceleration_and_parallelization
        """
        # default: fall back on autodetection
        res = None

        optarch = build_option('optarch') or ''
        # take into account that optarch value is a dictionary if it is specified by compiler family
        if isinstance(optarch, dict):
            comp_fam = self.toolchain.comp_family()
            optarch = optarch.get(comp_fam, '')
        optarch = optarch.upper()

        # The list of GMX_SIMD options can be found
        # http://manual.gromacs.org/documentation/2018/install-guide/index.html#simd-support
        if 'MIC-AVX512' in optarch and LooseVersion(self.version) >= LooseVersion('2016'):
            res = 'AVX_512_KNL'
        elif 'AVX512' in optarch and LooseVersion(self.version) >= LooseVersion('2016'):
            res = 'AVX_512'
        elif 'AVX2' in optarch and LooseVersion(self.version) >= LooseVersion('5.0'):
            res = 'AVX2_256'
        elif 'AVX' in optarch:
            res = 'AVX_256'
        elif 'SSE3' in optarch or 'SSE2' in optarch or 'MARCH=NOCONA' in optarch:
            # Gromacs doesn't have any GMX_SIMD=SSE3 but only SSE2 and SSE4.1 [1].
            # According to [2] the performance difference between SSE2 and SSE4.1 is minor on x86
            # and SSE4.1 is not supported by AMD Magny-Cours[1].
            res = 'SSE2'
        elif optarch == OPTARCH_GENERIC:
            cpu_arch = get_cpu_architecture()
            if cpu_arch == X86_64:
                res = 'SSE2'
            else:
                res = 'None'
        elif optarch:
            warn_msg = "--optarch configuration setting set to %s but not taken into account; " % optarch
            warn_msg += "compiling GROMACS for the current host architecture (i.e. the default behavior)"
            self.log.warning(warn_msg)
            print_warning(warn_msg)

        if res:
            self.log.info("Target architecture based on optarch configuration option ('%s'): %s", optarch, res)
        else:
            self.log.info("No target architecture specified based on optarch configuration option ('%s')", optarch)

        return res

    def is_double_precision_cuda_build(self):
        """Check if the current build step involves double precision and CUDA"""
        cuda = get_software_root('CUDA')
        return cuda and self.double_prec_pattern in self.cfg['configopts']

    def prepare_step(self, *args, **kwargs):
        """Custom prepare step for GROMACS."""

        # With the intel toolchain the -ftz build flag is automatically added, causing
        # denormal results being flushed to zero. This will cause errors for very small
        # arguments without FMA support since some intermediate results might be denormal.
        # [https://redmine.gromacs.org/issues/2335]
        # Set -fp-model precise on non-FMA CPUs to produce correct results.
        if self.toolchain.comp_family() == toolchain.INTELCOMP:
            cpu_features = get_cpu_features()
            if 'fma' not in cpu_features:
                self.log.info("FMA instruction not supported by this CPU: %s", cpu_features)
                self.log.info("Setting precise=True intel toolchain option to remove -ftz build flag")
                self.toolchain.options['precise'] = True

        # This must be called after enforcing the precise option otherwise the
        # change will be ignored.
        super(EB_GROMACS, self).prepare_step(*args, **kwargs)

    def configure_step(self):
        """Custom configuration procedure for GROMACS: set configure options for configure or cmake."""

        if LooseVersion(self.version) >= LooseVersion('4.6'):
            cuda = get_software_root('CUDA')
            if cuda:
                # CUDA with double precision is currently not supported in GROMACS yet
                # If easyconfig explicitly have double_precision=True error out,
                # otherwise warn about it and skip the double precision build
                if self.cfg.get('double_precision'):
                    raise EasyBuildError("Double precision is not available for GPU build. " +
                                         "Please explicitly set \"double_precision = False\" " +
                                         "or remove it in the easyconfig file.")
                if self.double_prec_pattern in self.cfg['configopts']:
                    if self.cfg.get('double_precision') is None:
                        # Only print warning once when trying double precision
                        # build the first time
                        self.cfg['double_precision'] = False
                        self.log.info("Double precision is not available for " +
                                      "GPU build. Skipping the double precision build.")

                    self.log.info("skipping configure step")
                    return

                if LooseVersion(self.version) >= LooseVersion('2021'):
                    self.cfg.update('configopts', "-DGMX_GPU=CUDA -DCUDA_TOOLKIT_ROOT_DIR=%s" % cuda)
                else:
                    self.cfg.update('configopts', "-DGMX_GPU=ON -DCUDA_TOOLKIT_ROOT_DIR=%s" % cuda)

                # Set CUDA capabilities based on template value.
                if '-DGMX_CUDA_TARGET_SM' not in self.cfg['configopts']:
                    cuda_cc_semicolon_sep = self.cfg.get_cuda_cc_template_value(
                        "cuda_cc_semicolon_sep").replace('.', '')
                    self.cfg.update('configopts', '-DGMX_CUDA_TARGET_SM="%s"' % cuda_cc_semicolon_sep)
            else:
                # explicitly disable GPU support if CUDA is not available,
                # to avoid that GROMACS finds and uses a system-wide CUDA compiler
                self.cfg.update('configopts', "-DGMX_GPU=OFF")

        # PLUMED detection
        # enable PLUMED support if PLUMED is listed as a dependency
        # and PLUMED support is either explicitly enabled (plumed = True) or unspecified ('plumed' not defined)
        plumed_root = get_software_root('PLUMED')
        if self.cfg['plumed'] and not plumed_root:
            msg = "PLUMED support has been requested but PLUMED is not listed as a dependency."
            raise EasyBuildError(msg)
        elif plumed_root and self.cfg['plumed'] is False:
            self.log.info('PLUMED was found, but compilation without PLUMED has been requested.')
            plumed_root = None

        if plumed_root:
            self.log.info('PLUMED support has been enabled.')

            # Need to check if PLUMED has an engine for this version
            engine = 'gromacs-%s' % self.version

            (out, _) = run_cmd("plumed-patch -l", log_all=True, simple=False)
            if not re.search(engine, out):
                plumed_ver = get_software_version('PLUMED')
                msg = "There is no support in PLUMED version %s for GROMACS %s: %s" % (plumed_ver, self.version, out)
                if self.cfg['ignore_plumed_version_check']:
                    self.log.warning(msg)
                else:
                    raise EasyBuildError(msg)

            # PLUMED patching must be done at different stages depending on
            # version of GROMACS. Just prepare first part of cmd here
            plumed_cmd = "plumed-patch -p -e %s" % engine

        # Ensure that the GROMACS log files report how the code was patched
        # during the build, so that any problems are easier to diagnose.
        # The GMX_VERSION_STRING_OF_FORK feature is available since 2020.
        if (LooseVersion(self.version) >= LooseVersion('2020') and
                '-DGMX_VERSION_STRING_OF_FORK=' not in self.cfg['configopts']):
            gromacs_version_string_suffix = 'EasyBuild-%s' % EASYBUILD_VERSION
            if plumed_root:
                gromacs_version_string_suffix += '-PLUMED-%s' % get_software_version('PLUMED')
            self.cfg.update('configopts', '-DGMX_VERSION_STRING_OF_FORK=%s' % gromacs_version_string_suffix)

        if LooseVersion(self.version) < LooseVersion('4.6'):
            self.log.info("Using configure script for configuring GROMACS build.")

            if self.cfg['build_shared_libs']:
                self.cfg.update('configopts', "--enable-shared --disable-static")
            else:
                self.cfg.update('configopts', "--enable-static")

            # Use external BLAS and LAPACK
            self.cfg.update('configopts', "--with-external-blas --with-external-lapack")
            env.setvar('LIBS', "%s %s" % (os.environ['LIBLAPACK'], os.environ['LIBS']))

            # Don't use the X window system
            self.cfg.update('configopts', "--without-x")

            # OpenMP is not supported for versions older than 4.5.
            if LooseVersion(self.version) >= LooseVersion('4.5'):
                # enable OpenMP support if desired
                if self.toolchain.options.get('openmp', None):
                    self.cfg.update('configopts', "--enable-threads")
                else:
                    self.cfg.update('configopts', "--disable-threads")
            elif self.toolchain.options.get('openmp', None):
                raise EasyBuildError("GROMACS version %s does not support OpenMP" % self.version)

            # GSL support
            if get_software_root('GSL'):
                self.cfg.update('configopts', "--with-gsl")
            else:
                self.cfg.update('configopts', "--without-gsl")

            # actually run configure via ancestor (not direct parent)
            self.cfg['configure_cmd'] = "./configure"
            ConfigureMake.configure_step(self)

            # Now patch GROMACS for PLUMED between configure and build
            if plumed_root:
                run_cmd(plumed_cmd, log_all=True, simple=True)

        else:
            if '-DGMX_MPI=ON' in self.cfg['configopts']:
                mpi_numprocs = self.cfg.get('mpi_numprocs', 0)
                if mpi_numprocs == 0:
                    self.log.info("No number of test MPI tasks specified -- using default: %s",
                                  self.cfg['parallel'])
                    mpi_numprocs = self.cfg['parallel']

                elif mpi_numprocs > self.cfg['parallel']:
                    self.log.warning("Number of test MPI tasks (%s) is greater than value for 'parallel': %s",
                                     mpi_numprocs, self.cfg['parallel'])

                mpiexec = self.cfg.get('mpiexec')
                if mpiexec:
                    mpiexec_path = which(mpiexec)
                    if mpiexec_path:
                        self.cfg.update('configopts', "-DMPIEXEC=%s" % mpiexec_path)
                        self.cfg.update('configopts', "-DMPIEXEC_NUMPROC_FLAG=%s" %
                                        self.cfg.get('mpiexec_numproc_flag'))
                        self.cfg.update('configopts', "-DNUMPROC=%s" % mpi_numprocs)
                    elif self.cfg['runtest']:
                        raise EasyBuildError("'%s' not found in $PATH", mpiexec)
                else:
                    raise EasyBuildError("No value found for 'mpiexec'")
                self.log.info("Using %s as MPI executable when testing, with numprocs flag '%s' and %s tasks",
                              mpiexec_path, self.cfg.get('mpiexec_numproc_flag'),
                              mpi_numprocs)

            if LooseVersion(self.version) >= LooseVersion('2019'):
                # Building the gmxapi interface requires shared libraries,
                # this is handled in the class initialisation so --module-only works
                self.cfg.update('configopts', "-DGMXAPI=ON")

                if LooseVersion(self.version) >= LooseVersion('2020'):
                    # build Python bindings if Python is loaded as a dependency
                    python_root = get_software_root('Python')
                    if python_root:
                        bin_python = os.path.join(python_root, 'bin', 'python')
                        self.cfg.update('configopts', "-DPYTHON_EXECUTABLE=%s" % bin_python)
                        self.cfg.update('configopts', "-DGMX_PYTHON_PACKAGE=ON")

            # Now patch GROMACS for PLUMED before cmake
            if plumed_root:
                if LooseVersion(self.version) >= LooseVersion('5.1'):
                    # Use shared or static patch depending on
                    # setting of self.cfg['build_shared_libs']
                    # and adapt cmake flags accordingly as per instructions
                    # from "plumed patch -i"
                    if self.cfg['build_shared_libs']:
                        mode = 'shared'
                    else:
                        mode = 'static'
                    plumed_cmd = plumed_cmd + ' -m %s' % mode

                run_cmd(plumed_cmd, log_all=True, simple=True)

            # prefer static libraries, if available
            if self.cfg['build_shared_libs']:
                self.cfg.update('configopts', "-DGMX_PREFER_STATIC_LIBS=OFF")
            else:
                self.cfg.update('configopts', "-DGMX_PREFER_STATIC_LIBS=ON")

            # always specify to use external BLAS/LAPACK
            self.cfg.update('configopts', "-DGMX_EXTERNAL_BLAS=ON -DGMX_EXTERNAL_LAPACK=ON")

            # disable GUI tools
            self.cfg.update('configopts', "-DGMX_X11=OFF")

            # convince to build for an older architecture than present on the build node by setting GMX_SIMD CMake flag
            # it does not make sense for Cray, because OPTARCH is defined by the Cray Toolchain
            if self.toolchain.toolchain_family() != toolchain.CRAYPE:
                gmx_simd = self.get_gromacs_arch()
                if gmx_simd:
                    if LooseVersion(self.version) < LooseVersion('5.0'):
                        self.cfg.update('configopts', "-DGMX_CPU_ACCELERATION=%s" % gmx_simd)
                    else:
                        self.cfg.update('configopts', "-DGMX_SIMD=%s" % gmx_simd)

            # set regression test path
            prefix = 'regressiontests'
            if any([src['name'].startswith(prefix) for src in self.src]):
                self.cfg.update('configopts', "-DREGRESSIONTEST_PATH='%%(builddir)s/%s-%%(version)s' " % prefix)

            # enable OpenMP support if desired
            if self.toolchain.options.get('openmp', None):
                self.cfg.update('configopts', "-DGMX_OPENMP=ON")
            else:
                self.cfg.update('configopts', "-DGMX_OPENMP=OFF")

            imkl_root = get_software_root('imkl')
            if imkl_root:
                # using MKL for FFT, so it will also be used for BLAS/LAPACK
                imkl_include = os.path.join(os.getenv('MKLROOT'), 'mkl', 'include')
                self.cfg.update('configopts', '-DGMX_FFT_LIBRARY=mkl -DMKL_INCLUDE_DIR="%s" ' % imkl_include)
                libs = os.getenv('LAPACK_STATIC_LIBS').split(',')
                mkl_libs = [os.path.join(os.getenv('LAPACK_LIB_DIR'), lib) for lib in libs if lib != 'libgfortran.a']
                mkl_libs = ['-Wl,--start-group'] + mkl_libs + ['-Wl,--end-group -lpthread -lm -ldl']
                self.cfg.update('configopts', '-DMKL_LIBRARIES="%s" ' % ';'.join(mkl_libs))
            else:
                for libname in ['BLAS', 'LAPACK']:
                    libdir = os.getenv('%s_LIB_DIR' % libname)
                    if self.toolchain.toolchain_family() == toolchain.CRAYPE:
                        libsci_mpi_mp_lib = glob.glob(os.path.join(libdir, 'libsci_*_mpi_mp.a'))
                        if libsci_mpi_mp_lib:
                            self.cfg.update('configopts', '-DGMX_%s_USER=%s' % (libname, libsci_mpi_mp_lib[0]))
                        else:
                            raise EasyBuildError("Failed to find libsci library to link with for %s", libname)
                    else:
                        # -DGMX_BLAS_USER & -DGMX_LAPACK_USER require full path to library
                        # prefer shared libraries when using FlexiBLAS-based toolchain
                        if self.toolchain.blas_family() == toolchain.FLEXIBLAS:
                            libs = os.getenv('%s_SHARED_LIBS' % libname).split(',')
                        else:
                            libs = os.getenv('%s_STATIC_LIBS' % libname).split(',')

                        libpaths = [os.path.join(libdir, lib) for lib in libs if not lib.startswith('libgfortran')]
                        self.cfg.update('configopts', '-DGMX_%s_USER="%s"' % (libname, ';'.join(libpaths)))
                        # if libgfortran.a is listed, make sure it gets linked in too to avoiding linking issues
                        if 'libgfortran.a' in libs:
                            env.setvar('LDFLAGS', "%s -lgfortran -lm" % os.environ.get('LDFLAGS', ''))

            # no more GSL support in GROMACS 5.x, see http://redmine.gromacs.org/issues/1472
            if LooseVersion(self.version) < LooseVersion('5.0'):
                # enable GSL when it's provided
                if get_software_root('GSL'):
                    self.cfg.update('configopts', "-DGMX_GSL=ON")
                else:
                    self.cfg.update('configopts', "-DGMX_GSL=OFF")

            # include flags for linking to zlib/XZ in $LDFLAGS if they're listed as a dep;
            # this is important for the tests, to correctly link against libxml2
            for dep, link_flag in [('XZ', '-llzma'), ('zlib', '-lz')]:
                root = get_software_root(dep)
                if root:
                    libdir = get_software_libdir(dep)
                    ldflags = os.environ.get('LDFLAGS', '')
                    env.setvar('LDFLAGS', "%s -L%s %s" % (ldflags, os.path.join(root, libdir), link_flag))

            # complete configuration with configure_method of parent
            out = super(EB_GROMACS, self).configure_step()

            # for recent GROMACS versions, make very sure that a decent BLAS, LAPACK and FFT is found and used
            if LooseVersion(self.version) >= LooseVersion('4.6.5'):
                patterns = [
                    r"Using external FFT library - \S*",
                    r"Looking for dgemm_ - found",
                    r"Looking for cheev_ - found",
                ]
                for pattern in patterns:
                    regex = re.compile(pattern, re.M)
                    if not regex.search(out):
                        raise EasyBuildError("Pattern '%s' not found in GROMACS configuration output.", pattern)

            # Make sure compilation of CPU detection code did not fail
            patterns = [
                r".*detection program did not compile.*",
            ]
            for pattern in patterns:
                regex = re.compile(pattern, re.M)
                if regex.search(out):
                    raise EasyBuildError("Pattern '%s' found in GROMACS configuration output.", pattern)

    def build_step(self):
        """
        Custom build step for GROMACS; Skip if CUDA is enabled and the current
        iteration is for double precision
        """

        if self.is_double_precision_cuda_build():
            self.log.info("skipping build step")
        else:
            super(EB_GROMACS, self).build_step()

    def test_step(self):
        """Run the basic tests (but not necessarily the full regression tests) using make check"""

        if self.is_double_precision_cuda_build():
            self.log.info("skipping test step")
        else:
            # allow to escape testing by setting runtest to False
            if self.cfg['runtest'] is None or self.cfg['runtest']:

                libdir = os.path.join(self.installdir, 'lib')
                libdir_backup = None

                if build_option('rpath'):
                    # temporarily copy 'lib' to installation directory when RPATH linking is enabled;
                    # required to fix errors like:
                    #     "ImportError: libgmxapi.so.0: cannot open shared object file: No such file or directory"
                    # occurs with 'make test' because _gmxapi.*.so only includes %(installdir)/lib in RPATH section,
                    # while the libraries are only there after install step...

                    # keep in mind that we may be performing an iterated installation:
                    # if there already is an existing 'lib' dir in the installation,
                    # we temporarily move it out of the way (and then restore it after running the tests)
                    if os.path.exists(libdir):
                        libdir_backup = find_backup_name_candidate(libdir)
                        self.log.info("%s already exists, moving it to %s while running tests...",
                                      libdir, libdir_backup)
                        shutil.move(libdir, libdir_backup)

                    copy_dir('lib', libdir)

                orig_runtest = self.cfg['runtest']
                # make very sure OMP_NUM_THREADS is set to 1, to avoid hanging GROMACS regression test
                env.setvar('OMP_NUM_THREADS', '1')

                if self.cfg['runtest'] is None or isinstance(self.cfg['runtest'], bool):
                    self.cfg['runtest'] = 'check'

                # run 'make check' or whatever the easyconfig specifies
                # in parallel since it involves more compilation
                self.cfg.update('runtest', "-j %s" % self.cfg['parallel'])
                super(EB_GROMACS, self).test_step()

                if build_option('rpath'):
                    # clean up temporary copy of 'lib' in installation directory,
                    # this was only there to avoid ImportError when running the tests before populating
                    # the installation directory
                    remove_dir(libdir)

                    if libdir_backup:
                        self.log.info("Restoring %s to %s after running tests", libdir_backup, libdir)
                        shutil.move(libdir_backup, libdir)

                self.cfg['runtest'] = orig_runtest

    def install_step(self):
        """
        Custom install step for GROMACS; figure out where libraries were installed to.
        """
        # Skipping if CUDA is enabled and the current iteration is double precision
        if self.is_double_precision_cuda_build():
            self.log.info("skipping install step")
        else:
            # run 'make install' in parallel since it involves more compilation
            self.cfg.update('installopts', "-j %s" % self.cfg['parallel'])
            super(EB_GROMACS, self).install_step()

    def extensions_step(self, fetch=False):
        """ Custom extensions step, only handle extensions after the last iteration round"""
        if self.iter_idx < self.variants_to_build - 1:
            self.log.info("skipping extension step %s", self.iter_idx)
        else:
            # Reset installopts etc for the benefit of the gmxapi extension
            self.cfg['install_cmd'] = self.orig_install_cmd
            self.cfg['build_cmd'] = self.orig_build_cmd
            self.cfg['installopts'] = self.orig_installopts
            # Set runtest to None so that the gmxapi extension doesn't try to
            # run "check" as a command
            orig_runtest = self.cfg['runtest']
            self.cfg['runtest'] = None
            super(EB_GROMACS, self).extensions_step(fetch)
            self.cfg['runtest'] = orig_runtest

    def get_lib_subdir(self):
        # the GROMACS libraries get installed in different locations (deeper subdirectory),
        # depending on the platform;
        # this is determined by the GNUInstallDirs CMake module;
        # rather than trying to replicate the logic, we just figure out where the library was placed

        if LooseVersion(self.version) < LooseVersion('5.0'):
            libname = 'libgmx*.%s' % self.libext
        else:
            libname = 'libgromacs*.%s' % self.libext
        lib_subdir = None
        for libdir in ['lib', 'lib64']:
            if os.path.exists(os.path.join(self.installdir, libdir)):
                for subdir in [libdir, os.path.join(libdir, '*')]:
                    libpaths = glob.glob(os.path.join(self.installdir, subdir, libname))
                    if libpaths:
                        lib_subdir = os.path.dirname(libpaths[0])[len(self.installdir) + 1:]
                        self.log.info("Found lib subdirectory that contains %s: %s", libname, lib_subdir)
                        break
        if not lib_subdir:
            raise EasyBuildError("Failed to determine lib subdirectory in %s", self.installdir)

        return lib_subdir

    def make_module_req_guess(self):
        """Custom library subdirectories for GROMACS."""
        guesses = super(EB_GROMACS, self).make_module_req_guess()
        if not self.lib_subdir:
            try:
                self.lib_subdir = self.get_lib_subdir()
            except EasyBuildError as error:
                if build_option('force') and build_option('module_only'):
                    self.log.info("No lib subdirectory directory found in installation: %s", error)
                    self.log.info("You are forcing module creation for a non-existent installation!")
                else:
                    raise error
        guesses.update({
            'LD_LIBRARY_PATH': [self.lib_subdir],
            'LIBRARY_PATH': [self.lib_subdir],
            'PKG_CONFIG_PATH': [os.path.join(self.lib_subdir, 'pkgconfig')],
        })
        return guesses

    def sanity_check_step(self):
        """Custom sanity check for GROMACS."""

        dirs = [os.path.join('include', 'gromacs')]

        # in GROMACS v5.1, only 'gmx' binary is there
        # (only) in GROMACS v5.0, other binaries are symlinks to 'gmx'
        # bins/libs that never have an _mpi suffix
        bins = []
        libnames = []
        # bins/libs that may have an _mpi suffix
        mpi_bins = []
        mpi_libnames = []
        if LooseVersion(self.version) < LooseVersion('5.1'):
            mpi_bins.extend(['mdrun'])

        if LooseVersion(self.version) >= LooseVersion('5.0'):
            mpi_bins.append('gmx')
            mpi_libnames.append('gromacs')
        else:
            bins.extend(['editconf', 'g_lie', 'genbox', 'genconf'])
            libnames.extend(['gmxana'])
            if LooseVersion(self.version) >= LooseVersion('4.6'):
                if self.cfg['build_shared_libs']:
                    mpi_libnames.extend(['gmx', 'md'])
                else:
                    libnames.extend(['gmx', 'md'])
            else:
                mpi_libnames.extend(['gmx', 'md'])

            if LooseVersion(self.version) >= LooseVersion('4.5'):
                if LooseVersion(self.version) >= LooseVersion('4.6'):
                    if self.cfg['build_shared_libs']:
                        mpi_libnames.append('gmxpreprocess')
                    else:
                        libnames.append('gmxpreprocess')
                else:
                    mpi_libnames.append('gmxpreprocess')

        # also check for MPI-specific binaries/libraries
        if self.toolchain.options.get('usempi', None):
            if LooseVersion(self.version) < LooseVersion('4.6'):
                mpisuff = self.cfg.get('mpisuffix', '_mpi')
            else:
                mpisuff = '_mpi'

            mpi_bins.extend([binary + mpisuff for binary in mpi_bins])
            mpi_libnames.extend([libname + mpisuff for libname in mpi_libnames])

        suffixes = ['']

        # make sure that configopts is a list:
        configopts_list = self.cfg['configopts']
        if isinstance(configopts_list, str):
            configopts_list = [configopts_list]

        lib_files = []
        bin_files = []

        dsuff = None
        if not get_software_root('CUDA'):
            for configopts in configopts_list:
                # add the _d suffix to the suffix, in case of double precision
                if self.double_prec_pattern in configopts:
                    dsuff = '_d'

        if dsuff:
            suffixes.extend([dsuff])

        lib_files.extend([
            'lib%s%s.%s' % (x, suff, self.libext) for x in libnames + mpi_libnames for suff in suffixes
        ])
        bin_files.extend([b + suff for b in bins + mpi_bins for suff in suffixes])

        if not self.lib_subdir:
            self.lib_subdir = self.get_lib_subdir()

        # pkgconfig dir not available for earlier versions, exact version to use here is unclear
        if LooseVersion(self.version) >= LooseVersion('4.6'):
            dirs.append(os.path.join(self.lib_subdir, 'pkgconfig'))

        custom_paths = {
            'files': [os.path.join('bin', b) for b in bin_files] +
            [os.path.join(self.lib_subdir, lib) for lib in lib_files],
            'dirs': dirs,
        }
        super(EB_GROMACS, self).sanity_check_step(custom_paths=custom_paths)

    def run_all_steps(self, *args, **kwargs):
        """
        Put configure options in place for different variants, (no)mpi, single/double precision.
        """
        # Save installopts so we can reset it later. The gmxapi pip install
        # can't handle the -j argument.
        self.orig_installopts = self.cfg['installopts']

        # keep track of config/build/installopts specified in easyconfig
        # file, so we can include them in each iteration later
        common_config_opts = self.cfg['configopts']
        common_build_opts = self.cfg['buildopts']
        common_install_opts = self.cfg['installopts']

        self.orig_install_cmd = self.cfg['install_cmd']
        self.orig_build_cmd = self.cfg['build_cmd']

        self.cfg['configopts'] = []
        self.cfg['buildopts'] = []
        self.cfg['installopts'] = []

        if LooseVersion(self.version) < LooseVersion('4.6'):
            prec_opts = {
                'single': '--disable-double',
                'double': '--enable-double',
            }
            mpi_type_opts = {
                'nompi': '--disable-mpi',
                'mpi': '--enable-mpi'
            }
        else:
            prec_opts = {
                'single': '-DGMX_DOUBLE=OFF',
                'double': '-DGMX_DOUBLE=ON',
            }
            mpi_type_opts = {
                'nompi': '-DGMX_MPI=OFF -DGMX_THREAD_MPI=ON',
                'mpi': '-DGMX_MPI=ON -DGMX_THREAD_MPI=OFF'
            }

        # Double precision pattern so search for in configopts
        self.double_prec_pattern = prec_opts['double']

        # For older versions we only build/install the mdrun part for
        # the MPI variant. So we need to be able to specify the
        # install target depending on variant.
        self.cfg['install_cmd'] = 'make'
        if LooseVersion(self.version) < LooseVersion('5'):
            # Use the fact that for older versions we just need to
            # build and install mdrun for the MPI part
            build_opts = {
                'nompi': '',
                'mpi': 'mdrun'
            }
            install_opts = {
                'nompi': 'install',
                'mpi': 'install-mdrun'
            }
        else:
            build_opts = {
                'nompi': '',
                'mpi': ''
            }
            install_opts = {
                'nompi': 'install',
                'mpi': 'install'
            }

        precisions = ['single']
        if self.cfg.get('double_precision') is None or self.cfg.get('double_precision'):
            precisions.append('double')

        mpitypes = ['nompi']
        if self.toolchain.options.get('usempi', None):
            mpitypes.append('mpi')

        # We need to count the number of variations to build.
        versions_built = []
        # Handle the different variants
        for precision in precisions:
            for mpitype in mpitypes:
                versions_built.append('%s precision %s' % (precision, mpitype))
                var_confopts = []
                var_buildopts = []
                var_installopts = []

                var_confopts.append(mpi_type_opts[mpitype])
                var_confopts.append(prec_opts[precision])
                if LooseVersion(self.version) < LooseVersion('4.6'):
                    suffix = ''
                    if mpitype == 'mpi':
                        suffix = "--program-suffix={0}".format(self.cfg.get('mpisuffix', '_mpi'))
                        if precision == 'double':
                            suffix += '_d'
                    var_confopts.append(suffix)

                var_buildopts.append(build_opts[mpitype])
                var_installopts.append(install_opts[mpitype])

                self.cfg.update('configopts', ' '.join(var_confopts + [common_config_opts]))
                self.cfg.update('buildopts', ' '.join(var_buildopts + [common_build_opts]))
                self.cfg.update('installopts', ' '.join(var_installopts + [common_install_opts]))
        self.variants_to_build = len(self.cfg['configopts'])

        self.log.debug("List of configure options to iterate over: %s", self.cfg['configopts'])
        self.log.info("Building these variants of GROMACS: %s", ', '.join(versions_built))
        return super(EB_GROMACS, self).run_all_steps(*args, **kwargs)

        self.cfg['install_cmd'] = self.orig_install_cmd
        self.cfg['build_cmd'] = self.orig_build_cmd

        self.log.info("A full regression test suite is available from the GROMACS web site: %s", self.cfg['homepage'])



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for SHRiMP, implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
"""
import glob
import os
import shutil

import easybuild.tools.environment as env
from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.tools.build_log import EasyBuildError


class EB_SHRiMP(ConfigureMake):
    """Support for building SHRiMP."""

    def configure_step(self):
        """Add openmp compilation flag to CXX_FLAGS."""

        cxxflags = os.getenv('CXXFLAGS')

        env.setvar('CXXFLAGS', "%s %s" % (cxxflags, self.toolchain.get_flag('openmp')))

    def install_step(self):
        """Install SHRiMP by copying files to install dir, and fix permissions."""

        try:
            for d in ["bin", "utils"]:
                shutil.copytree(d, os.path.join(self.installdir, d))

            cwd = os.getcwd()

            os.chdir(os.path.join(self.installdir, 'utils'))
            for f in glob.glob("*.py"):
                self.log.info("Fixing permissions of %s in utils" % f)
                os.chmod(f, 0o755)

            os.chdir(cwd)

        except OSError as err:
            raise EasyBuildError("Failed to copy files to install dir: %s", err)

    def sanity_check_step(self):
        """Custom sanity check for SHRiMP."""

        custom_paths = {
            'files': ['bin/%s' % x for x in ['fasta2fastq', 'gmapper', 'mergesam',
                                             'prettyprint', 'probcalc', 'probcalc_mp',
                                             'shrimp2sam', 'shrimp_var']],
            'dirs': ['utils'],
        }

        super(EB_SHRiMP, self).sanity_check_step(custom_paths=custom_paths)

    def make_module_req_guess(self):
        """Add both 'bin' and 'utils' directories to PATH."""

        guesses = super(EB_SHRiMP, self).make_module_req_guess()

        guesses.update({'PATH': ['bin', 'utils']})

        return guesses

    def make_module_extra(self):
        """Set SHRIMP_FOLDER environment variable in module."""

        txt = super(EB_SHRiMP, self).make_module_extra()

        txt += self.module_generator.set_environment('SHRIMP_FOLDER', self.installdir)

        return txt



##
# This file is an EasyBuild reciPY as per https://github.com/easybuilders/easybuild
#
# Copyright:: Copyright 2012-2023 Uni.Lu/LCSB, NTUA
# Authors::   Cedric Laczny <cedric.laczny@uni.lu>, Fotis Georgatos <fotis@cern.ch>, Kenneth Hoste
# License::   MIT/GPL
# $Id$
#
# This work implements a part of the HPCBIOS project and is a component of the policy:
# http://hpcbios.readthedocs.org/en/latest/HPCBIOS_2012-94.html
##
"""
EasyBuild support for building and installing Bowtie2, implemented as an easyblock

@author: Cedric Laczny (Uni.Lu)
@author: Fotis Georgatos (Uni.Lu)
@author: Kenneth Hoste (Ghent University)
"""
from distutils.version import LooseVersion
import os

from easybuild.easyblocks.generic.makecp import MakeCp
from easybuild.framework.easyconfig import CUSTOM


class EB_Bowtie2(MakeCp):
    """
    Support for building bowtie2 (ifast and sensitive read alignment)
    - create Make.UNKNOWN
    - build with make and install
    """
    @staticmethod
    def extra_options(extra_vars=None):
        """Change default values of options"""
        extra = MakeCp.extra_options()
        # files_to_copy is not mandatory here
        extra['files_to_copy'][2] = CUSTOM
        return extra

    def __init__(self, *args, **kwargs):
        """Bowtie2 easyblock constructor, define class variables."""
        super(EB_Bowtie2, self).__init__(*args, **kwargs)

        self.bowtie2_files = ['bowtie2', 'bowtie2-build', 'bowtie2-inspect', 'MANUAL', 'MANUAL.markdown', 'NEWS']
        if LooseVersion(self.version) >= LooseVersion('2.2.0'):
            self.bowtie2_files.extend(['bowtie2-align-l', 'bowtie2-align-s', 'bowtie2-build-l', 'bowtie2-build-s',
                                       'bowtie2-inspect-l', 'bowtie2-inspect-s'])
        else:
            self.bowtie2_files.extend(['bowtie2-align'])

        if LooseVersion(self.version) >= LooseVersion('2.0.5'):
            self.bowtie2_files.append('LICENSE')

    def configure_step(self):
        """No custom configuration step for Bowtie2"""
        pass

    def build_step(self):
        """Build Bowtie2, make sure right compilation flags are used"""

        cc = os.getenv('CC')
        if cc:
            self.cfg.update('buildopts', 'CC="%s"' % cc)

        cxx = os.getenv('CXX')
        if cxx:
            self.cfg.update('buildopts', 'CPP="%s" CXX="%s"' % (cxx, cxx))

        cxxflags = os.getenv('CXXFLAGS')
        if cxxflags:
            self.cfg.update('buildopts', 'RELEASE_FLAGS="%s"' % cxxflags)

        super(EB_Bowtie2, self).build_step()

    def install_step(self):
        """
        Install by copying files to install dir
        """
        self.cfg['files_to_copy'] = [(self.bowtie2_files, 'bin'), 'doc', 'example', 'scripts']
        super(EB_Bowtie2, self).install_step()

    def sanity_check_step(self):
        """Custom sanity check for Bowtie2."""
        custom_paths = {
            'files': [os.path.join('bin', f) for f in self.bowtie2_files],
            'dirs': ['doc', 'example', 'scripts']
        }
        super(EB_Bowtie2, self).sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for Go packages, implemented as an EasyBlock

@author: Pavel Grochal (INUITS)
"""
import os
from distutils.version import LooseVersion

import easybuild.tools.environment as env
from easybuild.framework.easyblock import EasyBlock
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.modules import get_software_root, get_software_version
from easybuild.tools.run import run_cmd


class GoPackage(EasyBlock):
    """Builds and installs a Go package, and provides a dedicated module file."""

    @staticmethod
    def extra_options(extra_vars=None):
        """Easyconfig parameters specific to Go packages."""
        if extra_vars is None:
            extra_vars = {}
        extra_vars.update({
            'modulename': [None, "Module name of the Go package, when building non-native module", CUSTOM],
            'forced_deps': [None, "Force specific version of Go package, when building non-native module", CUSTOM],
        })
        return extra_vars

    def prepare_step(self, *args, **kwargs):
        """Go-specific preparations."""
        super(GoPackage, self).prepare_step(*args, **kwargs)

        if get_software_root('Go') is None:
            raise EasyBuildError("Failed to pick go command to use. Is it listed in dependencies?")

        if LooseVersion(get_software_version('Go')) < LooseVersion("1.11"):
            raise EasyBuildError("Go version < 1.11 doesn't support installing modules from go.mod")

    def configure_step(self):
        """Configure Go package build/install."""

        # enforce use of go modules
        env.setvar('GO111MODULE', 'on', verbose=False)
        # set bin folder
        env.setvar('GOBIN', os.path.join(self.installdir, 'bin'), verbose=False)

        # creates log entries for go being used, for debugging
        run_cmd("go version", verbose=False, trace=False)
        run_cmd("go env", verbose=False, trace=False)

    def build_step(self):
        """If Go package is not native go module, lets try to make the module."""

        go_mod_file = 'go.mod'
        go_sum_file = 'go.sum'

        if not os.path.exists(go_mod_file) or not os.path.isfile(go_mod_file):
            self.log.warn("go.mod not found! This is not natively supported go module. Trying to init module.")

            if self.cfg['modulename'] is None:
                raise EasyBuildError("Installing non-native go module. You need to specify 'modulename' in easyconfig")

            # for more information about migrating to go modules
            # see: https://blog.golang.org/migrating-to-go-modules

            # go mod init
            cmd = ' '.join(['go', 'mod', 'init', self.cfg['modulename']])
            run_cmd(cmd, log_all=True, simple=True)

            if self.cfg['forced_deps']:
                for dep in self.cfg['forced_deps']:
                    # go get specific dependencies which locks them in go.mod
                    cmd = ' '.join(['go', 'get', '%s@%s' % dep])
                    run_cmd(cmd, log_all=True, simple=True)

            # note: ... (tripledot) used below is not a typo, but go wildcard pattern
            # which means: anything you can find in this directory, including all subdirectories
            # see: 'go help packages' or https://golang.org/pkg/cmd/go/internal/help/
            # see: https://stackoverflow.com/a/28031651/2047157

            # building and testing will add packages to go.mod
            run_cmd('go build ./...', log_all=True, simple=True)
            run_cmd('go test ./...', log_all=True, simple=True)

            # tidy up go.mod
            run_cmd('go mod tidy', log_all=True, simple=True)

            # build and test again, to ensure go mod tidy didn't removed anything needed
            run_cmd('go build ./...', log_all=True, simple=True)
            run_cmd('go test ./...', log_all=True, simple=True)

            self.log.warn('Include generated go.mod and go.sum via patch to ensure locked dependencies '
                          'and run this easyconfig again.')
            run_cmd('cat go.mod', log_all=True, simple=True)
            run_cmd('cat go.sum', log_all=True, simple=True)

        if not os.path.exists(go_sum_file) or not os.path.isfile(go_sum_file):
            raise EasyBuildError("go.sum not found! This module has no locked dependency versions.")

    def install_step(self):
        """Install Go package to a custom path"""

        # actually install Go package
        cmd = ' '.join([
            self.cfg['preinstallopts'],
            'go',
            'install',
            # print commands as they are executed,
            # including downloading and installing of package deps as listed in the go.mod file
            '-x',
            self.cfg['installopts'],
        ])
        run_cmd(cmd, log_all=True, log_ok=True, simple=True)

    def sanity_check_step(self):
        """Custom sanity check for Go package."""

        # Check if GoPackage produced binary and can run help on it
        custom_paths = {
            'files': ['bin/%s' % self.name.lower()],
            'dirs': [],
        }
        custom_commands = ['%s --help' % self.name.lower()]

        super(GoPackage, self).sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)
